#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging

from dremio.models.flight_config import FlightConfigDict
from .version import *  # for __version__, __author__

__all__ = ["Dremio"]

from ._mixins import (
    BaseClass,
    _MixinCatalog,
    _MixinSQL,
    _MixinQuery,
    _MixinDataset,
    _MixinFolder,
    _MixinFlight,
    _MixinReflection,
    _MixinUser,
    _MixinRole,
)


# TODO: write a few lines about mixins
class Dremio(
    _MixinRole,
    _MixinUser,
    _MixinQuery,
    _MixinFlight,
    _MixinFolder,
    _MixinDataset,
    _MixinReflection,
    _MixinSQL,
    _MixinCatalog,
    BaseClass,
):
    """This class is the main class of the Dremio connector. [learn more](https://github.com/continental/pydremio/blob/master/docs/DREMIO_METHODS.md)"""

    @staticmethod
    def from_env(
        hostname: str = "DREMIO_HOSTNAME",
        username: str = "DREMIO_USERNAME",
        password: str = "DREMIO_PASSWORD",
        modus: str = "DREMIO_MODUS",
        port: str = "DREMIO_PORT",
        protocol: str = "DREMIO_PROTOCOL",
        *,
        flight_port: str = "DREMIO_FLIGHT_PORT",
        flight_tls: str = "DREMIO_FLIGHT_TLS",
        flight_path_to_certs: str = "DREMIO_FLIGHT_PATH_TO_CERTS",
        load_dotenv: bool = False,
    ) -> "Dremio":
        """Create a new Dremio instance from environment variables. [learn more](https://github.com/continental/pydremio/blob/master/docs/DREMIO_LOGIN.md)
        The environment variables are:
          - DREMIO_HOSTNAME
          - DREMIO_USERNAME
          - DREMIO_PASSWORD
          - DREMIO_MODUS
          - DREMIO_PORT
          - DREMIO_PROTOCOL
          - DREMIO_FLIGHT_PORT
          - DREMIO_FLIGHT_TLS
          - DREMIO_FLIGHT_PATH_TO_CERTS


        Parameters:
          hostname: environment variable name for the hostname.
          username: environment variable name for the username.
          password: environment variable name for the password.
          modus: environment variable name for the modus.
          port: environment variable name for the port.
          protocol: environment variable name for the protocol.
          load_dotenv: load the dotenv file. Default: False.

        Returns:
          Dremio: Dremio instance.
        """
        if load_dotenv:
            from dotenv import load_dotenv as ldotenv
            ldotenv()
        import os

        hostname = os.getenv(hostname) or os.getenv("DREMIO_ENDPOINT")  # type: ignore
        username = os.getenv(username)  # type: ignore
        password = os.getenv(password)  # type: ignore
        modus = os.getenv(modus) or "stand_alone"
        port = os.getenv(port) or "443"
        protocol = os.getenv(protocol) or "https"
        if not hostname:
            raise ValueError("No hostname provided")
        if not username or not password:
            logging.warning("No username or password provided")
        if modus not in ["stand_alone", "server"]:
            raise ValueError("Modus must be either stand_alone or server")
        try:
            int(port)
        except:
            raise ValueError("Port must be a valid integer")
        if protocol not in ["https", "http"]:
            raise ValueError("Protocol must be either https or http")

        flight_conf: FlightConfigDict = {}
        if f_port := os.getenv(flight_port):
            flight_conf["port"] = int(f_port)
        if f_tls := os.getenv(flight_tls):
            flight_conf["tls"] = bool(f_tls)
        if f_path_to_certs := os.getenv(flight_path_to_certs):
            flight_conf["path_to_certs"] = str(f_path_to_certs)

        return Dremio(hostname, username=username, password=password, modus=modus, port=int(port), protocol=protocol, flight_config=flight_conf)  # type: ignore


__all__ = ["_MixinSQL"]  # this is like `export ...` in typescript
import logging
from typing import overload
import requests
from time import sleep

from ..utils.multithreading import ThreadWithReturnValue as Thread

from ..utils.converter import to_dict, path_to_dotted
from ..models import *
from ..models.jobs import SQLType

from . import BaseClass
from ._catalog import _MixinCatalog


class _MixinSQL(_MixinCatalog, BaseClass):

    def _http_query(self, sql_request: Union[str, SQLRequest]) -> UUID:
        """Execute SQL query on Dremio.
        Please use `Dremio.start_job(...)` if you simply run a sql query.
        This Method is for low level implementation.

        Args:
            sql_request (Union[str, SQLRequest]): SQL statement as string or SQLRequest object

        Returns:
            UUID: job id
        """
        url = f"{self.hostname}/api/v3/sql"
        if isinstance(sql_request, str):
            sql_request = SQLRequest(sql_request)
        payload = to_dict(sql_request)
        response = requests.post(url, headers=self._headers, json=payload)
        self._raise_error(response)
        return JobId(**response.json()).id

    @overload
    def _http_query_result(
        self,
        sql_request: Union[str, SQLRequest],
        *,
        limit: Union[int, None] = None,
        offset: int = 0,
        timeout_in_sec: int = 1000,
        return_exception: Literal[False] = False,
    ) -> tuple[JobResult, Job]:
        """Execute a SQL request and get the results.
        If you need the Exception as return value, set `return_exception=True`.

        Parameters:
          sql_request: SQL request as SQLRequest object.
          limit: The limit of the results rows for pagination.
          offset: The offset of the results rows pagination.
          timeout_in_sec: seconds until TimeoutError raising
          return_exception: overload this method to return the exception instead of raising it. Default: False

        Raises:
          DremioError

        Returns:
          tuple[JobResult|None, Job]: The job results and the job.
        """
        ...

    @overload
    def _http_query_result(
        self,
        sql_request: Union[str, SQLRequest],
        *,
        limit: Union[int, None] = None,
        offset: int = 0,
        timeout_in_sec: int = 1000,
        return_exception: Literal[True] = True,
    ) -> tuple[Union[JobResult, None], Union[Job, None], Union[Exception, None]]:
        """Execute a SQL request and get the results. This method don't raise errors.
        Overloaded!!!

        Parameters:
          sql_request: SQL request as SQLRequest object.
          limit: The limit of the results rows for pagination.
          offset: The offset of the results rows pagination.
          timeout_in_sec: seconds until TimeoutError raising
          return_exception: overload this method to return the exception instead of raising it.

        Returns:
          tuple[JobResult|None, Job]: The job results and the job.
        """
        ...

    def _http_query_result(
        self,
        sql_request: Union[str, SQLRequest],
        *,
        limit: Union[int, None] = None,
        offset: int = 0,
        timeout_in_sec: int = 1000,
        return_exception: Literal[True, False] = False,
    ):
        if return_exception:
            # only for special cases, when you need only the job and the error, but not the result
            # please use sql_result for normal cases
            result: Union[JobResult, None] = None
            error: Union[Exception, None] = None
            job: Union[Job, None] = None
            try:
                job_id = self._http_query(sql_request)
                result = self._http_wait_for_job_result(
                    job_id, limit, offset, timeout_in_sec
                )
            except Exception as e:
                error = e
            try:
                job = self.get_job_info(job_id)
            except Exception as e:
                error = e
                job = None
            return result, job, error
        else:
            job_id = self._http_query(sql_request)
            result = self._http_wait_for_job_result(
                job_id, limit, offset, timeout_in_sec
            )
            job = self.get_job_info(job_id)
            return result, job

    def _http_start_job(self, sql_request: Union[str, SQLRequest]) -> Job:
        """Start a sql job over http.
        This method is for low level implementations.

        Args:
            sql_request (Union[str, SQLRequest]): sql query as string or SQLRequest object.

        Returns:
            Job: job object of the started job
        """
        job_id = self._http_query(sql_request=sql_request)
        return self.get_job_info(job_id)

    def _http_start_job_on_dataset(
        self, id: Union[UUID, str], limit: Union[int, None] = None
    ) -> Job:
        """Start a job on a dataset. Is necessary for Dremio.Dataset.run()

        Args:
            id (Union[UUID, str]): dataset id
            limit (Union[int, None], optional): row limit for request. Defaults to None.

        Raises:
            ValueError: DremioError

        Returns:
            UUID: job id
        """
        item = self.get_catalog_by_id(id)
        if type(item) != Dataset:
            raise ValueError("requested resource is not dataset")
        q = f"SELECT * FROM {path_to_dotted(item.path)}"
        if limit:
            q += " LIMIT {limit}"
        sql_req = SQLRequest(q)
        job_id = self._http_query(sql_req)
        return self.get_job_info(job_id)

    def _http_dataset_result(
        self,
        dataset: Dataset | list[str] | str,
        limit: Union[int, None] = None,
        offset: int = 0,
        timeout_in_sec: int = 1000,
    ) -> tuple[JobResult, Job]:
        """Get the result of a dataset."""
        if isinstance(dataset, list) or isinstance(dataset, str):
            dataset = cast(Dataset, self._get_catalog_object(path=dataset))
        if not isinstance(dataset, Dataset):
            raise TypeError("Dataset is needed to run a dataset")
        if not dataset.sql:
            raise DremioConnectorError(
                "No SQL statement in dataset",
                f"please check the sql statement in dataset {dataset.id}",
            )
        return self._http_query_result(
            SQLRequest(dataset.sql, dataset.sqlContext),
            limit=limit,
            offset=offset,
            timeout_in_sec=timeout_in_sec,
        )

    def get_job_info(self, id: Union[UUID, str]) -> Job:
        """Get the job info.

        Parameters:
          id: job uuid.

        Returns:
          Job: job object.
        """
        url = f"{self.hostname}/api/v3/job/{str(id)}"
        response = requests.get(url, headers=self._headers)
        self._raise_error(response)
        job = cast(Job, response.json())
        job.id = id  # type: ignore
        return job

    def cancel_job(self, id: Union[UUID, str]) -> Job:
        """Cancel the job.

        Parameters:
          id: job uuid.

        Returns:
          Job: job object.
        """
        url = f"{self.hostname}/api/v3/job/{str(id)}/cancel"
        response = requests.post(url, headers=self._headers)
        self._raise_error(response)
        # check job status
        job: Job = self.get_job_info(id)
        if job.jobState != "CANCELED":
            raise DremioError(
                "Job not canceled", f"Job {id} still in state {job.jobState}"
            )
        return job

    def _http_get_job_result(
        self, id: Union[UUID, str], limit: Union[int, None] = None, offset: int = 0
    ) -> JobResult:
        """ONLY FOR LOW LEVEL IMPLEMENTATIONS! Use Dremio.query() otherwise!
        Get the job results. DID NOT WAIT FOR THE RESULTS! Use `wait_for_job_result` instead.

        Parameters:
          id: The job id.
          limit: The limit of the results rows for pagination.
          offset: The offset of the results rows pagination.

        Returns:
          JobResult: The job results.
        """
        url = f"{self.hostname}/api/v3/job/{str(id)}/results?offset={offset}"
        if limit:
            url += f"&limit={limit}"
        response = requests.get(url, headers=self._headers)
        self._raise_error(response)
        result = JobResult.from_dict(response.json())
        if offset != 0 or limit:
            return result

        # paginate through all result rows
        threads: list[Thread] = []
        for i in range(len(result.rows), result.rowCount, 500):
            x = Thread(target=self._http_get_job_result, args=(id, 500, i))
            threads.append(x)
            x.start()
        for thread in threads:
            r: Union[JobResult, None] = thread.join()
            if not r:
                raise Exception(
                    f"Thread failed: {id}[{i}:{min(i+500,result.rowCount)}]"
                )
            result.rows += Schema[SQLType](r.rows)
        return result

    def _http_wait_for_job_result(
        self,
        id: Union[UUID, str],
        limit: Union[int, None] = None,
        offset: int = 0,
        timeout_in_sec: int = 500,
    ) -> JobResult:
        """Waits for the job results and give them back.
        **Attention**: this method can raise errors.

        Parameters:
          id: The job id.
          limit: The limit of the results rows for pagination.
          offset: The offset of the results rows pagination.
          timeout_in_sec: seconds until TimeoutError raising

        Returns:
          JobResult: The job results.
        """
        for ms in [50, 100, 100, 250, 250, 500, 750] + timeout_in_sec * [1000]:
            job: Job = self.get_job_info(id)
            if job.jobState in ["COMPLETED", "CANCELED", "FAILED"]:
                break
            sleep(ms / 1000)
        else:
            raise TimeoutError(f"Job {job.id} still {job.jobState}")
        if job.jobState == "COMPLETED":
            return self._http_get_job_result(id, limit, offset)
        else:
            raise DremioError(
                f"Job {job.id} {job.jobState} {job.errorMessage}",
                f"{job.cancellationReason}",
            )

    @property
    def http(self) -> "HTTP":
        """Namespace: Collection of methods related to data fetching via http.
        It's highly recommended to use Dremio.flight instead!!!

        Returns:
            HTTP: Namespace of http data fetching methods
        """
        return HTTP(self)


class HTTP:
    """Namespace: Collection of methods related to data fetching via http."""

    def __init__(self, mixin: _MixinSQL):
        self.query = mixin._http_query_result
        self.start_job = mixin._http_start_job
        self.query_dataset = mixin._http_dataset_result



from dataclasses import dataclass, field
from uuid import UUID
from typing import Optional, Literal, Union
from datetime import datetime

from .custom_objects import CatalogElement
from .dremio_utils import Config, MetadataPolicy

from .accesscontrol import (
    Owner,
    AccessControlList,
    Privilege,
)


@dataclass
class Home:
    entityType: Literal["home"]
    id: UUID
    name: str
    tag: str
    children: Optional[list[CatalogElement]] = None


@dataclass
class Space:
    entityType: Literal["space"]
    id: UUID
    name: str
    tag: str
    createdAt: datetime
    children: Optional[list[CatalogElement]] = None
    owner: Optional[Owner] = None
    accessControlList: Optional[AccessControlList] = None
    permissions: Optional[list[Privilege]] = None


@dataclass
class Function:
    id: UUID
    path: list[str]
    tag: str
    type: Literal["CONTAINER"]
    createdAt: datetime


@dataclass
class Source:
    entityType: Literal["source"]
    id: UUID
    tag: str
    type: str
    name: str
    createdAt: datetime
    metadataPolicy: MetadataPolicy
    accelerationGracePeriodMs: int
    accelerationRefreshPeriodMs: int
    accelerationNeverRefresh: bool
    accelerationNeverExpire: bool
    allowCrossSourceSelection: bool
    disableMetadataValidityCheck: bool
    config: Optional[Config] = None
    checkTableAuthorizer: Optional[bool] = None
    children: Optional[list[CatalogElement]] = None
    accessControlList: Optional[AccessControlList] = None
    permissions: Optional[list[Privilege]] = None
    owner: Optional[Owner] = None



from dataclasses import dataclass
from typing import Optional
from uuid import UUID

from .custom_objects import CatalogElement


@dataclass
class Lineage:
    sources: list[CatalogElement]
    parents: list[CatalogElement]
    children: list[CatalogElement]



import pytest, time

from tests.testutils.helper import random_name
from src.dremio import Dremio, DremioError, Space, Folder, Dataset


def test_folder_copy(
    dremio: Dremio, folderset: tuple[Folder, list[Dataset]], space: Space
):
    pytest.skip()
    # folder = folderset[0]
    # assert folder.children

    # folder_copy = folder.copy([space.name, random_name()])
    # assert folder_copy.children
    # assert len(folder.children) == len(folder_copy.children)

    # with pytest.raises(DremioError): # folder already exists
    #   folder.copy([space.name, folder_copy.name])

    # original_sql_statements = [ds.sql for ds in folderset[1]]
    # for child in folder_copy.children:
    #   ds = dremio.get_dataset(id=child.id)
    #   assert ds.sql in original_sql_statements



__all__ = ["_MixinFolder"]  # this is like `export ...` in typescript
import logging
import re
import requests

from ..utils.converter import path_to_list, path_to_dotted
from ..models import *

from . import BaseClass
from ._catalog import _MixinCatalog
from ._dataset import _MixinDataset


class _MixinFolder(_MixinDataset, _MixinCatalog, BaseClass):
    def get_folder(
        self,
        path: Union[list[str], str, None] = None,
        *,
        id: Union[UUID, str, None] = None,
    ) -> Folder:
        """Get the folder. [learn more](https://github.com/continental/pydremio/blob/master/docs/FOLDER.md)
        \nCould raise a DremioError if the folder does not exist or you have no permissions.

        Parameters:
          path: folder path: "A.B.C" or ["A","B","C"].
          id: folder uuid.

        Returns:
          Folder: folder object.
        """
        data = self._get_catalog_object(id=id, path=path)
        if data["entityType"] != "folder":
            raise TypeError("requested resource is not a folder")
        obj = cast(Folder, data)
        if type(obj) != Folder:
            raise TypeError("requested resource is not a folder")
        obj._dremio = self  # type: ignore
        return obj

    # Holgers implementation with loop
    def create_folder(self, path: Union[str, list[str]]) -> Folder:
        """Create a folder and all parent folders if not exist.

        Parameters:
          path: The path of the folder.
        """
        # create folder(s) (if don't exist)
        if isinstance(path, str):
            path = path_to_list(path)
        target_path_temp = []
        for p in path[:-1]:
            target_path_temp.append(p)
            new_folder = NewFolder(target_path_temp)
            try:
                self.create_catalog_item(new_folder)
            except DremioError as e:
                if e.status_code in [400, 409]:  # folder already exists
                    continue
                raise e
        self.create_catalog_item(NewFolder(path))  # can raise DremioError
        return self.get_folder(path)

    def delete_folder(
        self, path: Union[str, list[str]], recursive: bool = True
    ) -> bool:
        """Delete the folder.
        \nCould raise a TypeError if there is no folder under this path.
        \nCould raise a DremioError if the folder does not exist or you have no permissions.

        Parameters:
          path: The path of the folder.
          recursive: Delete all children. Default: True. If False, it will raise an DremioConnectorError if the folder is not empty.
        """
        if isinstance(path, str):
            path = path_to_list(path)
        folder = self.get_folder(path)
        folder_has_children = len(folder.children) > 0 if folder.children else False
        if not recursive and folder_has_children:
            raise DremioError("Folder not empty", f"Folder at {path} is not empty")
        return self.delete_catalog_item(folder.id)

    def _duplicate_folder(
        self,
        source_path: Union[list[str], str],
        target_path: Union[list[str], str],
        method: Literal["ref", "copy"],
        assume_privileges: bool = True,
        relative_references: bool = False,
        overwrite_existing: bool = False,
        _datasets_to_commit: list[Dataset] = [],
    ) -> Folder:
        source_path = path_to_list(source_path)
        target_path = path_to_list(target_path)
        try:
            source = self.get_folder(source_path)
        except DremioError as e:
            if e.status_code == 404:
                raise DremioError(
                    "Folder not found", f"Folder at {source_path} not found"
                )
            raise e
        if not source:
            raise DremioError("Folder not found", f"Folder at {source_path} not found")
        try:
            target = self.create_folder(target_path)
        except DremioError as e:
            if e.status_code != 409:  # folder already exists
                raise e
            target = self.get_folder(target_path)
        if assume_privileges:  # only on folders
            try:
                target.accessControlList = source.accessControlList
                target.commit()
            except DremioError as e:
                raise DremioError(
                    "Error updating folder",
                    f"Error updating folder {target_path}: {e.errorMessage}",
                )
        for child in source:
            new_child_path = target_path + [child.path[-1]]
            if child.type == "CONTAINER" and child.containerType == "FOLDER":
                if child.path == source_path or child.path == target_path:
                    continue
                self._duplicate_folder(
                    child.path,
                    new_child_path,
                    method,
                    assume_privileges,
                    relative_references,
                    overwrite_existing,
                    _datasets_to_commit,
                )
            elif child.type == "DATASET":
                if method == "copy":
                    ds: Dataset
                    try:
                        ds = self.copy_dataset(
                            child.path,
                            new_child_path,
                            assume_privileges=assume_privileges,
                            overwrite_existing=False,
                        )  # we don't use the existing overwrite flag to implement custom logging
                    except DremioError as e:
                        # warning is needed in multiple cases - made little helper function for it:
                        log_warning = lambda e: logging.warning(
                            f"Failed to copy {path_to_dotted(child.path)} - {e} ({self._get_url_of_object(child.path)})"
                        )

                        if e.status_code == 409:
                            if not overwrite_existing:
                                logging.info(
                                    f"{path_to_dotted(child.path)} already exists and is left as it is"
                                )
                                continue
                            try:
                                logging.info(
                                    f"{path_to_dotted(child.path)} exists and will be overwritten"
                                )
                                self.delete_dataset(new_child_path)
                                ds = self.copy_dataset(
                                    child.path,
                                    new_child_path,
                                    assume_privileges=assume_privileges,
                                    overwrite_existing=False,
                                )
                            except Exception as e:
                                log_warning(e)  # use warning helper lambda
                        else:
                            log_warning(e)  # use warning helper lambda
                            continue
                    # update sql references
                    if relative_references:
                        new_sql = re.sub(
                            r"\"?"
                            + r"\"?\.\"?".join(source_path)
                            + r"\"?",  # find 'Application.TEST_FOLDER' and '"Application"."TEST_FOLDER"'
                            path_to_dotted(target_path),
                            ds.sql or "",
                        )
                        ds.sql = new_sql
                        # add this dataset to a list to commit it later
                        # this is for the case that the datasets, that been used in this SQL statement, are not copied until now.
                        # if we commit after wards, we can face this problem
                        _datasets_to_commit.append(
                            ds
                        )  # has to be committed after wards

                elif method == "ref":
                    try:
                        self.reference_dataset(
                            child.path,
                            new_child_path,
                            assume_privileges=assume_privileges,
                            overwrite_existing=False,
                        )
                    except DremioError as e:
                        log_warning = lambda e: logging.warning(
                            f"Failed to reference {path_to_dotted(child.path)} - {e} ({self._get_url_of_object(child.path)})"
                        )
                        if e.status_code == 409:
                            if not overwrite_existing:
                                logging.info(
                                    f"{path_to_dotted(child.path)} already exists and is left as it is"
                                )
                                continue
                            try:
                                logging.info(
                                    f"{path_to_dotted(child.path)} exists and will be overwritten"
                                )
                                self.delete_dataset(new_child_path)
                                ds = self.reference_dataset(
                                    child.path,
                                    new_child_path,
                                    assume_privileges=assume_privileges,
                                    overwrite_existing=False,
                                )
                            except Exception as e:
                                log_warning(e)
                        else:
                            log_warning(e)
                            continue
        return target

    def copy_folder(
        self,
        source_path: Union[list[str], str],
        target_path: Union[list[str], str],
        *,
        assume_privileges: bool = True,
        relative_references: bool = False,
        overwrite_existing: bool = False,
    ) -> Folder:
        """Copy the folder by path. The copy will point on the same source like the original.
        If the target path does not exist, it will create the folder. [learn more](https://github.com/continental/pydremio/blob/master/docs/DREMIO_METHODS.md#copy-a-folder)

        Parameters:
          source_path: source path of the folder.
          target_path: target path of the folder.
          assume_privileges: Assume privileges. Default: True.
          relative_references: Change the sql statements in datasets to the new path, if they are inside the copied scope -> [learn more](https://github.com/continental/pydremio/blob/master/docs/FOLDER.md#copy-a-folder). Default: False
          overwrite_existing: Overwrites the existing folders and datasets with the same path in the target folder. Default: False

        Returns:
          Folder: The copied folder.
        """
        datasets_to_commit: list[Dataset] = []
        target = self._duplicate_folder(
            source_path,
            target_path,
            "copy",
            assume_privileges,
            relative_references,
            overwrite_existing=overwrite_existing,
            _datasets_to_commit=datasets_to_commit,
        )
        for ds in datasets_to_commit:
            try:
                ds.commit()
            except Exception as e:
                logging.error(
                    f"Error while committing new relative sql statement to {path_to_dotted(ds.path)}.",
                    e,
                )
        return target

    def reference_folder(
        self,
        source_path: Union[list[str], str],
        target_path: Union[list[str], str],
        *,
        assume_privileges: bool = True,
        overwrite_existing: bool = False,
    ) -> Folder:
        """Copy the folder by path. All children will be referenced.
        If the target path does not exist, it will create the folder. [learn more](https://github.com/continental/pydremio/blob/master/docs/FOLDER.md#reference-a-folder)

        Parameters:
          source_path: source path of the folder.
          target_path: target path of the folder.
          assume_privileges: Assume privileges. Default: True.
          overwrite_existing: Overwrites the existing folders and datasets with the same path in the target folder. Default: False

        Returns:
          Folder: The referenced folder.
        """
        return self._duplicate_folder(
            source_path,
            target_path,
            "ref",
            assume_privileges,
            relative_references=False,
            overwrite_existing=overwrite_existing,
        )



import pytest
from tests.testutils.container import DremioContainer, InitUser

from src.dremio import Dremio, AuthError


@pytest.fixture(scope="module", autouse=True)
def module_user() -> InitUser:
    return InitUser(userName="test", password="Test123!")


@pytest.fixture(scope="module", autouse=True)
def module_container(module_user: InitUser):
    container = DremioContainer(module_user)
    yield container.start()
    container.stop()


def test_login_success(module_container: DremioContainer, module_user: InitUser):
    hostname = module_container.get_container_host_ip()
    port = module_container.get_exposed_port(9047)
    dremio = Dremio(hostname=hostname, port=int(port), protocol="http")
    me = dremio.login(username=module_user.userName, password=module_user.password)
    assert me.userName == module_user.userName


def test_login_incorrect(module_container: DremioContainer, module_user: InitUser):
    hostname = module_container.get_container_host_ip()
    port = module_container.get_exposed_port(9047)
    dremio = Dremio(hostname=hostname, port=int(port), protocol="http")
    with pytest.raises(AuthError, match="Login incorrect or user not found"):
        dremio.login(username=module_user.userName, password="123456789")



__all__ = ["FlightConfig", "FlightConfigDict"]


from dataclasses import asdict, dataclass, field
from typing import Any, Literal, Optional, TypedDict, overload

import certifi

from ..utils.parser import cut_scheme


@dataclass
class FlightConfig:
    """This config is to configure arrow flight queries to Dremio.

    Properties:
      port (int): Port of the grpc endpoint of Dremio. Default: 32010.
      tls (bool): Enable to use TLS. If `https` set as protocol for dremio, it is it is likely enabled.
      disable_certificate_verification: Disable to use certs from `path_to_certs`.
      allow_autoconfig (bool): Disable to prevent automatic tls setting.
      path_to_certs: Path to certs like cacert.pem. Default: certifi.where().
      engine: Routing engine for Dremio Cloud. Default: None.
    """

    port: int = 32010
    tls: bool = False
    disable_certificate_verification: bool = False
    allow_autoconfig: bool = True
    path_to_certs: str = certifi.where()
    headers: dict[str, Any] = field(default_factory=lambda: {})
    session_properties: dict[str, Any] = field(default_factory=lambda: {})
    engine: Optional[str] = None
    _scheme: Optional[str] = None

    @property
    def tls_root_certs(self) -> bytes | None:
        if not self.path_to_certs:
            return None
        with open(self.path_to_certs, "rb") as f:
            return f.read()

    @property
    def scheme(self) -> str:
        if self._scheme:
            return self._scheme
        p = "grpc"
        # if not self.disable_certificate_verification:
        #     p += "+ssl"
        if self.tls:
            p += "+tls"
        return p

    @scheme.setter
    def scheme(self, v: str | None):
        self._scheme = v

    def uri(self, hostname: str) -> str:
        return f"{self.scheme}://{cut_scheme(hostname)}:{self.port}"

    @overload
    def get_headers(
        self, headers: dict[str, Any] = {}, as_bytes: Literal[False] = False
    ) -> list[tuple[str, str]]: ...

    @overload
    def get_headers(
        self, headers: dict[str, Any] = {}, as_bytes: Literal[True] = True
    ) -> list[tuple[bytes, bytes]]: ...

    def get_headers(
        self, headers: dict[str, Any] = {}, as_bytes: bool = False
    ) -> list[tuple[bytes, bytes]] | list[tuple[str, str]]:
        if as_bytes:
            return [
                (k.encode("utf-8"), f"{v}".encode("utf-8"))
                for k, v in (self.session_properties | self.headers | headers).items()
            ]
        return [
            (k, f"{v}")
            for k, v in (self.session_properties | self.headers | headers).items()
        ]


class FlightConfigDict(TypedDict, total=False):
    port: int
    tls: bool
    disable_certificate_verification: bool
    allow_autoconfig: bool
    path_to_certs: str
    engine: str
    headers: dict[str, Any]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This file contains all needed elements for the test modules in the same folder
"""
from dremio.version import *  # for __version__, __author__

import pytest
from typing import Iterator
from src.dremio import Dremio, NewSpace, Space, Folder, Dataset
from tests.testutils.container import DremioContainer
from tests.testutils.helper import random_name


@pytest.fixture(scope="session")
def container():
    container = DremioContainer()
    yield container.start()
    container.stop()


@pytest.fixture(scope="session")
def sample_dataset_paths(container: DremioContainer) -> Iterator[list[str]]:
    yield container.add_sample_sources("NYC-weather.csv", "NYC-taxi-trips.csv")


@pytest.fixture(scope="module")
def space(container: DremioContainer) -> Iterator[Space]:
    space: Space = container.dremio.create_catalog_item(NewSpace(random_name()))  # type: ignore
    yield space
    container.dremio.delete_catalog_item(space.id)


@pytest.fixture(scope="module")
def dremio(container: DremioContainer) -> Dremio:
    return container.dremio


@pytest.fixture(scope="function", autouse=True)
def clear_space(dremio: Dremio, space: Space):
    try:
        if not space.children:
            return
        for child in space.children:
            dremio.delete_catalog_item(child.id)
    except Exception as e:
        pytest.fail(f"error while clearing {space.name}: {e}")


@pytest.fixture
def folder(dremio: Dremio, space: Space) -> Iterator[Folder]:
    folder = dremio.create_folder([space.name, random_name()])
    yield folder
    folder.delete(recursive=True)


@pytest.fixture
def dataset(
    dremio: Dremio, folder: Folder, sample_dataset_paths: list[str]
) -> Iterator[Dataset]:
    dataset = dremio.create_dataset(
        folder.path + [random_name()], f"SELECT * FROM {sample_dataset_paths[0]}"
    )  # created with testutils.conatiner
    yield dataset
    dataset.delete()


@pytest.fixture
def folderset(
    dremio: Dremio, folder: Folder, sample_dataset_paths: list[str]
) -> Iterator[tuple[Folder, list[Dataset]]]:
    datasets = [
        dremio.create_dataset(folder.path + [random_name()], f"SELECT * FROM {sample}")
        for sample in sample_dataset_paths
    ]  # created with testutils.conatiner
    references = [ds.reference(folder.path + [random_name()]) for ds in datasets]
    datasets.extend(references)
    yield dremio.get_folder(id=folder.id), datasets
    for ds in datasets:
        ds.delete()



from dataclasses import dataclass
import json
from .. import dremio
from typing import Any, Literal, Optional, Union
from .accesscontrol import Role, User
from ..utils.converter import clear_at, to_dict


@dataclass
class Dependency:
    needed_by: list[str]  # list of IDs
    resource: dict  # dataset

    def as_dict(self) -> dict:
        return to_dict(self)

    def json(self) -> str:
        return json.dumps(self.as_dict(), indent=2)


@dataclass
class Backup:
    tree: dict[str, Any]
    users: list[User]
    roles: list[Role]
    dependencies: dict[str, Dependency]

    @staticmethod
    def load(file_path: str) -> "Backup":
        ext = file_path.split(".")[-1]
        if ext == "json":
            with open(file_path, "r") as f:
                data = json.load(f)
        else:
            raise ValueError("Invalid file extension")
        return Backup(data=data)

    def __init__(
        self,
        tree: dict[str, Any] = {},
        users: list[User] = [],
        roles: list[Role] = [],
        dependencies: dict[str, Dependency] = {},
        *,
        data: Union[dict, Any] = None,
    ):
        # set default
        self.tree = {}
        self.users = []
        self.roles = []
        self.dependencies = {}

        # parse data as Backup or dict
        if data:
            if isinstance(data, dict):
                self._backup_from_dict(data)
            elif isinstance(data, Backup):
                self.tree = data.tree
                self.users = data.users
                self.roles = data.roles
                self.dependencies = data.dependencies
            else:
                raise TypeError("restore data arg is not a valid type")

        # make it possible to overwrite attrs
        if tree:
            self.tree = tree
        if users:
            self.users = users
        if roles:
            self.roles = roles
        if dependencies:
            self.dependencies = dependencies

    def _backup_from_dict(self, data: dict):
        if "users" not in data:
            data["users"] = []
        if "roles" not in data:
            data["roles"] = []
        try:
            self.tree = data["tree"]
            self.users = [User(**clear_at(user)) for user in data["users"]]
            self.roles = [Role(**role) for role in data["roles"]]
            self.dependencies = {}
            try:
                for path, dep in data["dependencies"].items():
                    try:
                        self.dependencies[path] = Dependency(
                            needed_by=dep["needed_by"], resource=dep["resource"]
                        )
                    except:
                        print(f" dep '{path}' not loadable")
                        continue
            except:
                print("dependencies can not be loaded")

        except:
            raise TypeError("Invalid data")

    def as_dict(self) -> dict:
        return to_dict(self)

    def json(self) -> str:
        return json.dumps(self.as_dict(), indent=2)

    def save(self, path: str):
        ext = path.split(".")[-1]
        if ext == "json":
            with open(path, "w") as f:
                f.write(self.json())
        else:
            raise ValueError("Invalid file extension")

    def restore(
        self, dremio: Any, target_path: Union[str, None] = None, retry_limit: int = 5
    ) -> "RestoreReport":
        if not target_path:
            return dremio.restore(self, retry_limit)
        else:
            return dremio.restore_on_path(self, target_path, retry_limit)


@dataclass
class RestoreReportObject:
    path: list[str]
    old_id: str
    new_id: str
    old_path: Optional[list[str]] = None


ReasonToFail = Literal["permission", "unknown"]


@dataclass
class Failed:
    reason: ReasonToFail
    data: RestoreReportObject


@dataclass
class RestoreReport:
    missing_dependencies: dict[str, list[str]]  # {node_path:[dep_path]}
    successful: list[RestoreReportObject]
    already_exists: list[RestoreReportObject]
    missing_users: list[User]
    failed: list[Failed]

    def __init__(self):
        self.missing_dependencies = {}
        self.successful = []
        self.already_exists = []
        self.missing_users = []
        self.failed = []

    def add_missing_dependency(self, node_path: str, dependency_path: str):
        if node_path in self.missing_dependencies:
            self.missing_dependencies[node_path].append(dependency_path)
        else:
            self.missing_dependencies[node_path] = []

    def add_successful(
        self,
        path: list[str],
        new_id: str,
        old_id: str = "",
        old_path: Union[list[str], None] = None,
    ):
        r = RestoreReportObject(
            path=path, old_id=old_id, new_id=new_id, old_path=old_path
        )
        self.successful.append(r)

    def add_already_exists(
        self, path: list[str], old_id: str, old_path: Union[list[str], None] = None
    ):
        r = RestoreReportObject(path=path, old_id=old_id, new_id="", old_path=old_path)
        self.already_exists.append(r)

    def add_failed(
        self,
        reason: ReasonToFail,
        path: list[str],
        old_id: str,
        old_path: Union[list[str], None] = None,
    ):
        r = RestoreReportObject(path=path, old_id=old_id, new_id="", old_path=old_path)
        self.failed.append(Failed(reason, r))

    def as_dict(self) -> dict:
        return to_dict(self)

    def json(self) -> str:
        return json.dumps(self.as_dict(), indent=2)

    def save(self, path: str):
        ext = path.split(".")[-1]
        if ext == "json":
            with open(path, "w") as f:
                f.write(self.json())
        else:
            raise ValueError("Invalid file extension")



from dataclasses import dataclass
from datetime import datetime
from uuid import UUID
from typing import Literal, Optional, Any, TypedDict, Union, List, Dict, overload
from .schema import Schema, SQLType, SchemaField

import polars as pl
import pyarrow


@dataclass
class ReflectionRelationship:
    datasetId: UUID
    reflectionId: UUID
    relationship: Literal["CONSIDERED", "MATCHED", "CHOSEN"]


@dataclass
class Acceleration:
    reflectionRelationships: list[ReflectionRelationship]


@dataclass
class Job:
    jobState: Literal["COMPLETED", "CANCELED", "FAILED", "RUNNING"]
    errorMessage: str
    startedAt: datetime
    queryType: Literal[
        "UI_RUN",
        "UI_PREVIEW",
        "UI_INTERNAL_PREVIEW",
        "UI_INTERNAL_RUN",
        "UI_EXPORT",
        "ODBC",
        "JDBC",
        "REST",
        "ACCELERATOR_CREATE",
        "ACCELERATOR_DROP",
        "UNKNOWN",
        "PREPARE_INTERNAL",
        "ACCELERATOR_EXPLAIN",
        "UI_INITIAL_PREVIEW",
    ]
    cancellationReason: str
    queueId: Optional[str] = None
    resourceSchedulingEndedAt: Optional[datetime] = None
    resourceSchedulingStartedAt: Optional[datetime] = None
    rowCount: Optional[int] = None
    queueName: Optional[str] = None
    endedAt: Optional[datetime] = None
    acceleration: Optional[Acceleration] = None
    id: Optional[UUID] = None


class JobResultDict(TypedDict):
    rowCount: int
    schema: list[SchemaField]
    rows: list[dict[str, Any]]


@dataclass
class JobResult:
    """This is a job result object.
    You can run `json()`, `to_polars()` or `to_pandas()` on the result to use the data.
    """

    rowCount: int
    schema: Schema[pyarrow.DataType] | Schema[SQLType]
    rows: list[dict[str, Any]]
    # jobId: Optional[UUID] = None

    @staticmethod
    def from_arrow_table(arrow_table: pyarrow.Table) -> "JobResult":
        schema_names: list[str] = arrow_table.schema.names
        schema_types: list[pyarrow.DataType] = arrow_table.schema.types
        schema = Schema[pyarrow.DataType](
            [{"name": n, "type": t} for n, t in zip(schema_names, schema_types)]
        )
        rows = arrow_table.to_pylist()
        rowCount = len(rows)
        return JobResult(rowCount=rowCount, schema=schema, rows=rows)

    @staticmethod
    def from_dict(d: JobResultDict) -> "JobResult":
        schema = Schema[SQLType](d["schema"])
        return JobResult(rowCount=d["rowCount"], rows=d["rows"], schema=schema)

    @property
    def dict(self) -> JobResultDict:
        return {
            "rowCount": self.rowCount,
            "schema": list(self.schema),
            "rows": self.rows,
        }

    def preview_table(self, number_of_rows: int = 5) -> str:
        """Returns a preview table as string

        Args:
            number_of_rows (int, optional): number of rows, that should be included. -1 for all rows. Defaults to 5.

        Returns:
            str: table as printable string
        """
        from prettytable import PrettyTable

        t = PrettyTable(self.schema.names)
        t.align = "l"
        t.add_row(self.schema.type_names)
        t.add_divider()
        for r in self.rows[:number_of_rows]:
            t.add_row(list(r.values()))
        if number_of_rows < self.rowCount:
            t.add_row(["..."] * len(self.schema.names))
        return t.get_string()

    def __str__(self) -> str:
        return (
            f"\nlen: {self.rowCount}\tcols: {len(self.schema.names)}\n\n"
            + self.preview_table()
            + f"\n\nhint: run `.to_polars()` or `.to_pandas()` for more options\n"
        )

    def __repr__(self) -> str:
        return self.preview_table(50)

    @overload
    def __getitem__(self, key: str) -> List[Any]:
        """Get a specific column as list

        Args:
            key (str): column name

        Returns:
            List[Any]: column content as list
        """

    @overload
    def __getitem__(self, key: int) -> Dict[str, Any]:
        """Get a specific row as dict.

        Args:
            key (int): row number

        Returns:
            Dict[str, Any]: row content as dict
        """

    @overload
    def __getitem__(self, key: slice) -> List[Dict[str, Any]]:
        """Get a specific row as dict.

        Args:
            key (int): row number

        Returns:
            Dict[str, Any]: row content as dict
        """

    def __getitem__(
        self, key: str | int | slice
    ) -> List[Any] | Dict[str, Any] | List[Dict[str, Any]]:
        if (
            not isinstance(key, int)
            and not isinstance(key, str)
            and not isinstance(key, slice)
        ):
            raise TypeError("key has to be int, str or slice")

        if isinstance(key, int) or isinstance(key, slice):
            return self.rows[key]

        if key not in self.schema.names:
            raise KeyError(f"Key '{key}' not found in schema.")
        index = self.schema.names.index(key)
        col = self.to_polars().to_series(index)
        return col.to_list()

    def __iter__(self):
        return iter(self.rows)

    def __next__(self):
        for row in self.rows:
            yield row

    def __len__(self):
        return self.rowCount

    def json(self, indent: int = 2, *args, **kwargs) -> str:
        """Convert result into a json string."""
        import json

        return json.dumps(self.dict, indent=indent, *args, **kwargs)

    def to_polars(self) -> pl.DataFrame:
        """Convert result into a polars data frame.

        [Polars Docs](https://docs.pola.rs/py-polars/html/reference/dataframe/index.html)
        """
        return pl.DataFrame(self.rows)

    def to_pandas(self):
        """Convert result into a pandas data frame, but we recommend to use `to_polars`

        [Pandas Docs](https://pandas.pydata.org/docs/getting_started/intro_tutorials/03_subset_data.html#how-do-i-select-specific-columns-from-a-dataframe)
        """
        import pandas as pd

        return pd.DataFrame(self.rows)



from dataclasses import dataclass
from typing import Optional, Any, Literal

from .custom_objects import AccessControlList
from .dremio_utils import MetadataPolicy
from ..utils.converter import path_to_list


@dataclass
class NewFolder:
    path: list[str]
    accessControlList: Optional[AccessControlList] = None
    entityType: Literal["folder"] = "folder"

    def __post_init__(self):
        # convert if path is a dotted string
        if isinstance(self.path, str):
            self.path = path_to_list(self.path)


@dataclass
class NewDataset:
    path: list[str]
    type: Literal["PHYSICAL_DATASET", "VIRTUAL_DATASET"]
    sql: str
    id: Optional[str] = None
    accessControlList: Optional[AccessControlList] = None
    entityType: Literal["dataset"] = "dataset"

    def __post_init__(self):
        self.path = path_to_list(self.path)


@dataclass
class NewSpace:
    name: str
    accessControlList: Optional[AccessControlList] = None
    entityType: Literal["space"] = "space"


@dataclass
class NewSource:
    type: str
    name: str
    tag: Optional[str]
    config: Optional[dict[str, Any]]
    metadataPolicy: Optional[MetadataPolicy]
    accelerationGracePeriodMs: Optional[int] = None
    accelerationRefreshPeriodMs: Optional[int] = None
    accelerationNeverExpire: Optional[bool] = None
    accelerationNeverRefresh: Optional[bool] = None
    allowCrossSourceSelection: Optional[bool] = None
    disableMetadataValidityCheck: Optional[bool] = None
    entityType: Literal["source"] = "source"


@dataclass
class SQLRequest:
    sql: str
    context: Optional[list[str]] = None



import pytest

from src.dremio.utils.validation import Validation


def test_hostname_https():
    hostname = "dremio.com"
    res = Validation().hostname(hostname, None, "https")
    assert res == "https://dremio.com"


def test_hostname_http():
    hostname = "dremio.com"
    res = Validation().hostname(hostname, None, "http")
    assert res == "http://dremio.com"


def test_hostname_http_port():
    hostname = "dremio.com"
    res = Validation().hostname(hostname, 80, "http")
    assert res == "http://dremio.com"


def test_hostname_https_port():
    hostname = "dremio.com"
    res = Validation().hostname(hostname, 443, "https")
    assert res == "https://dremio.com"


def test_hostname_https_port80():
    hostname = "dremio.com"
    res = Validation().hostname(hostname, 80, "https")
    assert res == "https://dremio.com:80"


def test_hostname_http_port443_0():
    hostname = "dremio.com"
    res = Validation().hostname(hostname, 443, "http")
    assert res == "http://dremio.com:443"


def test_hostname_http_port443_1():
    hostname = "https://dremio.com"
    res = Validation().hostname(hostname, 443, "http")
    assert res == "http://dremio.com:443"


def test_hostname_none_none():
    hostname = "http://dremio.com:443"
    res = Validation().hostname(hostname)
    assert res == "http://dremio.com:443"



from dataclasses import dataclass, field
from typing import Optional, Any, Literal, Union
from uuid import UUID


@dataclass
class DatasetAccelerationRefreshPolicy:
    refreshPeriodMs: int
    gracePeriodMs: int
    method: Literal["FULL", "INCREMENTAL"]
    accelerationNeverExpire: bool
    accelerationNeverRefresh: bool
    refreshField: Optional[str] = None


@dataclass
class Config:
    accessKey: str
    secure: bool
    externalBucketlist: list[str]
    rootPath: str
    enableAsync: bool
    compatibilityMode: bool
    isCachingEnabled: bool
    maxCacheSpacePct: int
    requesterPays: bool
    enableFileStatusCheck: bool
    defaultCtasFormat: str
    isPartitionInferenceEnabled: bool
    credentialType: str


@dataclass
class StatusMessage:
    level = Literal["INFO", "WARN", "ERROR"]
    message: str


@dataclass
class State:
    status: Literal["good", "bad", "warn"]
    suggestedUserAction: str
    messages: Optional[list[StatusMessage]] = None


@dataclass
class MetadataPolicy:
    authTTLMs: int
    namesRefreshMs: int
    datasetRefreshAfterMs: int
    datasetExpireAfterMs: int
    datasetUpdateMode: Literal["PREFETCH", "PREFETCH_QUERIED", "INLINE"]
    deleteUnavailableDatasets: bool
    autoPromoteDatasets: bool


@dataclass
class File:
    entityType: Literal["file"]
    id: UUID
    path: list[str]


@dataclass
class Tags:
    tags: list[str]
    version: Union[str, int]


@dataclass
class Wiki:
    text: str
    version: int = 0
    _dataset_id: Union[UUID, None] = None
    _dremio: Any = None

    def __str__(self) -> str:
        return self.text

    def __repr__(self) -> str:
        return self.text

    def commit(self) -> "Wiki":
        return self._dremio.set_wiki(self._dataset_id, self, True)


@dataclass
class JobId:
    id: UUID



from dataclasses import InitVar, dataclass, field
from optparse import Option
import logging
from uuid import UUID
from typing import Any, Iterable, Optional, Literal, Type, Union, List, cast
from typing_extensions import Self
from datetime import datetime
from time import sleep
import pyarrow as pa
import pandas as pd
import polars as pl

from .schema import SQLType, Schema, SchemaField

from ..models.reflections import NewReflection, Reflection

from ..exceptions import DremioConnectorError

from .custom_objects import DatasetFormat
from .dremio_utils import Wiki

from .accesscontrol import (
    Owner,
    AccessControlList,
)

from .jobs import JobResult
from .dremio_utils import DatasetAccelerationRefreshPolicy

from .baseclasses import DremioAccessible, DremioObject
from ..utils.converter import path_to_dotted


@dataclass
class TypeClass:
    name: str
    precision: Optional[int] = None
    scale: Optional[int] = None


@dataclass
class Field:
    name: str
    type: TypeClass

    def __str__(self) -> str:
        return f"{self.name} ({self.type.name})"


class Columns:
    __fields: list[Field]

    def __init__(self, fields: list[Field]):
        self.__fields = fields

    def __getattr__(self, name: str) -> str:
        for f in self.__fields:
            if f.name == name:
                return f.name
        raise AttributeError(f"{name} is not an attribute of Fields")

    def __getitem__(self, key: str) -> Field:
        for f in self.__fields:
            if f.name == key:
                return f
        raise KeyError(f"{key} is not in Fields")

    def __str__(self) -> str:
        return "\t".join([str(f) for f in self.__fields])


def cols_type(**fields) -> Type[Columns]:
    return cast(Type[Columns], type("DatasetColumns", (Columns,), fields))


@dataclass
class Dataset(DremioAccessible, DremioObject):
    """The Dremio dataset object is a representation of a dataset in Dremio and contains all needed meta data.
    To use the data behind this dataset run it with `.run()`.

    It's necessary to commit the dataset after changing any values.

    ```python
    dataset = dremio.get_dataset('a.b.c')
    dataset.sql = "SELECT * FROM a.b.d"
    dataset.commit()
    ```

    """

    entityType: Literal["dataset"]
    id: UUID
    type: Literal["PHYSICAL_DATASET", "VIRTUAL_DATASET"]
    path: list[str]
    createdAt: datetime
    tag: str
    fields: list[Field]
    sql: Optional[str] = None
    sqlContext: Optional[list[str]] = None
    owner: Optional[Owner] = None
    accelerationRefreshPolicy: Optional[DatasetAccelerationRefreshPolicy] = None
    format: Optional[DatasetFormat] = None
    approximateStatisticsAllowed: Optional[bool] = None
    accessControlList: Optional[AccessControlList] = None
    _col: InitVar[Columns] = Columns([])

    def __post_init__(self, col: Columns):
        DatasetColumns = cols_type(**{f.name: f.name for f in self.fields})
        self._col = DatasetColumns(self.fields)  # type: ignore

    @property
    def col(self):
        return self._col  # type: ignore

    def __str__(self) -> str:
        return f"Dataset(path = '{path_to_dotted(self.path)}', id = {self.id})"

    def __repr__(self) -> str:
        return f"Dataset(path = '{path_to_dotted(self.path)}', id = {self.id})"

    @property
    def name(self) -> str:
        return self.path[-1]

    def pull(self) -> Self:
        """Get the newest version of this dataset from dremio.
        Could be useful for multi-threading.
        This is the opposite of .commit()

        Raises:
            TypeError: raises if no dremio instance is connected

        Returns:
            Self: gives you itself for chaining
        """
        if not self._dremio:
            raise TypeError(
                "This object has no dremio instance!\nTry to commit the dataset directly in the dremio sdk"
            )
        self = self._dremio.get_dataset(id=self.id)
        return self

    def run(self, method: Literal["arrow", "http"] = "arrow") -> JobResult:
        """Run a job on the dataset and get the results. [learn more](https://github.com/continental/pydremio/blob/master/docs/DATASET.md#run)
        Per default it uses arrow flight to get the data.
        http is slower and should only be used if arrow is not available.
        You can run `json()`, `to_polars()` or `to_pandas()` on the result to use the data.

        Parameters:
          method (str): The method to run the job. Can be "arrow" or "http". Default is "arrow".

        Returns:
          JobResult: Result of the dataset object.
        """
        if method == "http":
            if not self._dremio:
                raise TypeError(
                    "This Dataset has no dremio instance!\nTry to run the job directly in the dremio sdk"
                )
            job_id = self._dremio._http_start_job_on_dataset(self.id).id
            while True:
                job = self._dremio.get_job_info(job_id)
                state = job.jobState
                if state == "COMPLETED":
                    break
                elif state == "FAILED":
                    raise DremioConnectorError(
                        f"Dataset {self.id} run failed...", f"{job}"
                    )
                elif state == "CANCELED":
                    raise DremioConnectorError(
                        f"Dataset {self.id} run canceled...",
                        f"{job.cancellationReason} {job}",
                    )
                else:
                    sleep(0.5)
            return self._dremio._http_get_job_result(job_id)
        elif method == "arrow":
            if not self._dremio:
                raise TypeError(
                    "This Dataset has no dremio instance!\nTry to run the job directly in the dremio sdk"
                )
            table: pa.Table = self._dremio._flight_query(
                f"SELECT * FROM {path_to_dotted(self.path)}"
            )
            return JobResult.from_arrow_table(table)
        else:
            raise ValueError("method must be 'arrow' or 'http'")

    def run_to_pandas(self) -> pd.DataFrame:
        """Get the results via arrow flight as pandas dataframe. [learn more](https://github.com/continental/pydremio/blob/master/docs/DATASET.md#run)
        Shortcut for `Dataset.run().to_pandas()`

        Returns:
          pd.DataFrame: Result of the dataset object as pandas dataframe.
        """
        if not self._dremio:
            raise TypeError(
                "This Dataset has no dremio instance!\nTry to run the job directly in the dremio sdk"
            )
        table: pa.Table = self._dremio._flight_query(
            f"SELECT * FROM {path_to_dotted(self.path)}"
        )
        return table.to_pandas()

    def run_to_polars(self) -> pl.DataFrame:
        """Get the results via arrow flight as polars dataframe. [learn more](https://github.com/continental/pydremio/blob/master/docs/DATASET.md#run)
        Shortcut for `Dataset.run().to_polars()`

        Returns:
          pl.DataFrame: Result of the dataset object as polars dataframe.
        """
        if not self._dremio:
            raise TypeError(
                "This Dataset has no dremio instance!\nTry to run the job directly in the dremio sdk"
            )
        table: pa.Table = self._dremio._flight_query(self.sql)
        return pl.DataFrame(table)

    def copy(
        self,
        target_path: Union[list[str], str],
        *,
        assume_privileges: bool = True,
        overwrite_existing: bool = False,
    ) -> "Dataset":
        """Copy the dataset to the target path. Same as `Dremio.copy_dataset(this_dataset_path, target_path)`"""
        if not self._dremio:
            raise TypeError(
                "This Dataset has no dremio instance!\nTry to copy the dataset directly in the dremio sdk"
            )
        return self._dremio.copy_dataset(
            self.path,
            target_path,
            assume_privileges=assume_privileges,
            overwrite_existing=overwrite_existing,
        )

    def reference(
        self,
        target_path: Union[list[str], str],
        *,
        assume_privileges: bool = True,
        overwrite_existing: bool = False,
    ) -> "Dataset":
        """Reference the dataset to the target path. Same as `Dremio.reference_dataset(this_dataset_path, target_path)`"""
        if not self._dremio:
            raise TypeError(
                "This Dataset has no dremio instance!\nTry to reference the dataset directly in the dremio sdk"
            )
        return self._dremio.reference_dataset(
            self.path,
            target_path,
            assume_privileges=assume_privileges,
            overwrite_existing=overwrite_existing,
        )

    def create_reflection(self, name: str, reflection: NewReflection) -> Reflection:
        """Create reflection for a Dataset.

        Args:
            name (str): Name for reflection.
            reflection (NewReflection): Reflections settings

        Returns:
            Reflection (Reflection): New reflection.
        """
        if not self._dremio:
            raise TypeError(
                "This Dataset has no dremio instance!\nTry to refresh the dataset directly in the dremio sdk"
            )
        return self._dremio.create_reflection(
            dataset_id=self.id, name=name, reflection=reflection
        )

    def create_recommended_reflections(
        self, type: Literal["ALL", "RAW", "AGG"] = "ALL"
    ) -> list[Reflection]:
        """Creating and Retrieving Reflection Recommendations for a Dataset.

        Args:
            type (Literal[&quot;ALL&quot;, &quot;RAW&quot;, &quot;AGG&quot;], optional): The type of reflection recommendations you want to create and retrieve. Defaults to "ALL".

        Returns:
            Reflections (list[Reflection]): List of all reflections of the dataset.
        """
        if not self._dremio:
            raise TypeError(
                "This Dataset has no dremio instance!\nTry to refresh the dataset directly in the dremio sdk"
            )
        return self._dremio.create_recommended_reflections(self.id, type)

    def get_reflections(self) -> list[Reflection]:
        """Returns all reflections of this dataset.

        Raises:
            TypeError: Raises if there is no related dremio instance

        Returns:
            list[Reflection]: List of all reflections of this dataset. Empty list if there is none.
        """
        if not self._dremio:
            raise TypeError(
                "This Dataset has no dremio instance!\nTry to refresh the dataset directly in the dremio sdk"
            )
        return self._dremio.get_reflections_from_dataset(self.path)

    @property
    def reflections(self) -> list[Reflection]:
        """All reflections of this dataset.

        Returns:
            list[Reflection]: List of all reflections of this dataset. Empty list if there is none.
        """
        return self.get_reflections()

    def refresh(self) -> None:
        """Refresh the dataset reflections. Same as `Dremio.refresh_dataset(this_dataset_path)`"""
        if not self._dremio:
            raise TypeError(
                "This Dataset has no dremio instance!\nTry to refresh the dataset directly in the dremio sdk"
            )
        self._dremio.refresh_dataset(self.path)
        self.pull()

    def delete_reflections(self) -> None:
        """Delete all reflections of this dataset."""
        if not self._dremio:
            raise TypeError(
                "This Dataset has no dremio instance!\nTry to refresh the dataset directly in the dremio sdk"
            )
        refs = self.get_reflections()
        for ref in refs:
            try:
                self._dremio.delete_reflection(ref.id)
            except Exception as e:
                logging.warning(e)
        return None

    def get_wiki(self) -> Wiki:
        """Get the wiki of the dataset."""
        if not self._dremio:
            raise TypeError(
                "This Dataset has no dremio instance!\nTry to get the wiki directly in the dremio sdk"
            )
        wiki = self._dremio.get_wiki(self.id)
        wiki._dataset_id = self.id
        wiki._dremio = self._dremio
        return wiki

    def set_wiki_text(self, text: str) -> Wiki:
        """Set the wiki text of the dataset."""
        if not self._dremio:
            raise TypeError(
                "This Dataset has no dremio instance!\nTry to set the wiki text directly in the dremio sdk"
            )
        wiki = self._dremio.get_wiki(self.id)
        wiki.text = text
        return wiki.commit()

    # maybe later: reason for not implementing this is that the wiki is not a property of the dataset and should be handled with the knowledge of the wiki object
    # @property
    # def wiki(self) -> Wiki:
    #   """Get the wiki of the dataset.
    #   """
    #   return self.get_wiki()

    # @wiki.setter
    # def wiki(self, wiki:Wiki|str):
    #   if not self._dremio:
    #     raise TypeError("This Dataset has no dremio instance!\nTry to set the wiki directly in the dremio sdk")
    #   if isinstance(wiki, str):
    #     self.set_wiki_text(wiki)
    #     return
    #   elif isinstance(wiki, Wiki):
    #     self.set_wiki_text(wiki.text)
    #     return
    #   raise TypeError("wiki must be a Wiki object or a string")

    def delete(self) -> None:
        """Delete the dataset. Same as `Dremio.delete_dataset(this_dataset_path)`."""
        if not self._dremio:
            raise TypeError(
                "This Dataset has no dremio instance!\nTry to delete the dataset directly in the dremio sdk"
            )
        success: bool = False
        try:
            success = self._dremio.delete_dataset(self.path)
        except Exception as e:
            raise e
        if not success:
            raise DremioConnectorError(
                "Dataset not deleted", f"Dataset {self.path} could not be deleted"
            )
        del self

    @property
    def schema(self) -> Schema[SQLType]:
        """Get schema like `JobResult.schema`. This is just a stream lined TypedDict for `.fields`

        Returns:
            Schema[SQLType]: `.fields` as TypedDict.
        """
        return Schema[SQLType](
            [
                SchemaField(name=f.name, type=SQLType(name=f.type.name))
                for f in self.fields
            ]
        )



from dataclasses import dataclass
from typing import Literal, Optional
from uuid import UUID
from enum import Enum


Privilege = Literal[
    "READ",
    "WRITE",
    "ALTER_REFLECTION",
    "SELECT",
    "ALTER",
    "VIEW_REFLECTION",
    "MODIFY",
    "MANAGE_GRANTS",
    "CREATE_TABLE",
    "DROP",
    "EXTERNAL_QUERY",
    "INSERT",
    "TRUNCATE",
    "DELETE",
    "UPDATE",
    "EXECUTE",
    "CREATE_SOURCE",
    "ALL",
]
SystemRoles = Literal["ADMIN", "PUBLIC"]


@dataclass
class AccessControl:
    id: UUID
    permissions: list[Privilege]


@dataclass
class AccessControlList:
    users: Optional[list[AccessControl]] = None
    roles: Optional[list[AccessControl]] = None

    def __post_init__(self):
        if self.users is None:
            self.users = []
        if self.roles is None:
            self.roles = []

    def set_access_for_user(self, user_id: UUID, privileges: list[Privilege]):
        if self.users is None:
            self.users = []
        for user in self.users:
            if user.id == user_id:
                user.permissions = privileges
                return
        self.users.append(AccessControl(id=user_id, permissions=privileges))

    def add_access_for_user(self, user_id: UUID, privileges: list[Privilege]):
        if self.users is None:
            self.users = []
        for user in self.users:
            if user.id == user_id:
                user.permissions += privileges
                return
        self.users.append(AccessControl(id=user_id, permissions=privileges))

    def remove_access_for_user(self, user_id: UUID):
        if self.users is None:
            self.users = []
        for user in self.users:
            if user.id == user_id:
                self.users.remove(user)
                return

    def set_access_for_role(self, role_id: UUID, privileges: list[Privilege]):
        if self.roles is None:
            self.roles = []
        for role in self.roles:
            if role.id == role_id:
                role.permissions = privileges
                return
        self.roles.append(AccessControl(id=role_id, permissions=privileges))

    def add_access_for_role(self, role_id: UUID, privileges: list[Privilege]):
        if self.roles is None:
            self.roles = []
        for role in self.roles:
            if role.id == role_id:
                role.permissions += privileges
                return
        self.roles.append(AccessControl(id=role_id, permissions=privileges))

    def remove_access_for_role(self, role_id: UUID):
        if self.roles is None:
            self.roles = []
        for role in self.roles:
            if role.id == role_id:
                self.roles.remove(role)
                return


@dataclass
class Owner:
    ownerId: UUID
    ownerType: str  # Literal['USER']


@dataclass
class Permissions:
    canUploadProfiles: bool
    canDownloadProfiles: bool
    canEmailForSupport: bool
    canChatForSupport: bool
    canViewAllJobs: bool
    canCreateUser: bool
    canCreateRole: bool
    canCreateSource: bool
    canUploadFile: bool
    canManageNodeActivity: bool
    canManageEngines: bool
    canManageQueues: bool
    canManageEngineRouting: bool
    canManageSupportSettings: bool


@dataclass
class Role:
    id: UUID
    name: str
    type: Literal["INTERNAL", "EXTERNAL", "SYSTEM"]
    roles: Optional[list[dict[str, str]]] = None
    memberCount: Optional[int] = None
    description: Optional[str] = None


# TODO: Add member control


@dataclass
class User:
    id: UUID
    name: str
    firstName: str
    lastName: str
    email: str
    tag: str
    active: bool
    type: Optional[Literal["EnterpriseUser", "User"]] = None
    roles: Optional[list[Role]] = None
    source: Optional[str] = None

    @staticmethod
    def dict_factory(x):
        change = {"type": "@type"}
        return {k: v if k not in change else change[k] for (k, v) in x}


@dataclass
class CurrentUser:
    token: str
    userName: str
    firstName: str
    lastName: str
    expires: int
    email: str
    userId: str
    admin: bool
    clusterId: str
    clusterCreatedAt: int
    version: str
    permissions: Permissions
    userCreatedAt: int



__all__ = ["_MixinCatalog"]  # this is like `export ...` in typescript

import copy
import requests

from ..utils.converter import to_dict
from ..models import *

from . import BaseClass


class _MixinCatalog(BaseClass):

    def get_catalog_by_id(self, id: Union[UUID, str]) -> CatalogObject:
        """Get the catalog by id.

        Parameters:
          id: catalog uuid.

        Returns:
          CatalogObject: The catalog.
        """
        obj = self._T(self._get_catalog_object(id=id))
        if type(obj) in [Dataset, Folder]:
            obj._dremio = self  # type: ignore
        return obj

    def get_catalog_by_path(self, path: Union[list[str], str]) -> CatalogObject:
        """Get the catalog by path.

        Parameters:
          path: catalog path: "A.B.C" or ["A","B","C"].

        Returns:
          CatalogObject: The catalog.
        """
        obj = self._T(self._get_catalog_object(path=path))
        if type(obj) in [Dataset, Folder]:
            obj._dremio = self  # type: ignore
        return obj

    def create_catalog_item(
        self, item: Union[NewCatalogObject, CatalogObject, dict]
    ) -> CatalogObject:
        """Create the catalog item.

        Parameters:
          item: item that should be created.

        Returns:
          CatalogObject: created catalog item from dremio.
        """
        url = f"{self.hostname}/api/v3/catalog"
        try:
            item = to_dict(item)
        except:
            pass
        if isinstance(item, dict):
            item = self._T(item, True)
        new_item = copy.deepcopy(item)
        new_item.id = None  # type: ignore
        payload = to_dict(new_item)
        response = requests.post(url, headers=self._headers, json=payload)
        self._raise_error(response)
        return self._T(response.json())

    def update_catalog_item(
        self, id: Union[UUID, str], item: Union[CatalogObject, dict]
    ) -> CatalogObject:
        """Set the catalog."""
        url = f"{self.hostname}/api/v3/catalog/{id}"
        if isinstance(item, dict):
            item = self._T(item)
        item.id = id if item.id != id else item.id  # type: ignore
        payload = to_dict(item)
        response = requests.put(url, headers=self._headers, json=payload)
        self._raise_error(response)
        return self._T(response.json())

    def update_catalog_item_by_path(
        self, path: list[str], item: Union[CatalogObject, dict]
    ) -> CatalogObject:
        """Set the catalog by path."""
        id = self.get_catalog_by_path(path).id
        return self.update_catalog_item(id, item)

    def delete_catalog_item(self, id: Union[UUID, str]) -> bool:
        """Delete the catalog item."""
        url = f"{self.hostname}/api/v3/catalog/{id}"
        response = requests.delete(url, headers=self._headers)
        self._raise_error(response)
        if response.status_code != 204:
            raise DremioError("unknown error", "", response.status_code)
        return True

    def _refresh_catalog(self, id: Union[UUID, str]) -> CatalogObject:
        """Refresh the reflections associated with the specified table.

        Args:
            id (Union[UUID, str]): catalog item id.

        Returns:
            CatalogObject: The catalog object.
        """
        url = f"{self.hostname}/api/v3/catalog/{str(id)}/refresh"
        response = requests.post(url, headers=self._headers)
        self._raise_error(response)
        print(response)
        return self.get_catalog_by_id(id)



from .unions import *
from ..exceptions import *
from .inputs import *
from .custom_objects import *
from .backup import *
from .jobs import *
from .utils import *
from .reflections import *
from .accesscontrol import *
from .dremio_objects import *
from .dremio_utils import *
from .lineage import *
from .dataset import Dataset
from .folder import Folder
from .flight_config import FlightConfig, FlightConfigDict
from .reflections import *



import functools
from typing import Callable
import logging


def deprecated(hint: str):
    def decorator(func: Callable):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            logging.warn(
                DeprecationWarning(f"{func.__name__} is deprecated. {hint}"),
                stacklevel=2,
            )
            func.__doc__ = (
                f"> :warning: `{func.__name__}` **is deprecated**: {hint}\n"
                + (func.__doc__ or "")
            )
            return func(*args, **kwargs)

        return wrapper

    return decorator


def deprecated_interface(func: Callable):
    def wrapper(*args, **kwargs):
        s = f"{func.__name__} is deprecated."
        logging.warning(DeprecationWarning(s), stacklevel=2)
        func.__doc__ = s + "\n" + (func.__doc__ or "")
        return func(*args, **kwargs)

    return wrapper



from .dremio_utils import *
from .dremio_objects import *
from .custom_objects import *
from .inputs import *
from .dataset import Dataset
from .folder import Folder
from typing import Any, List, Dict, Optional, Type, Union, NewType

## RESPONSES ##

CatalogObject = Union[
    Source,
    Dataset,
    Space,
    Folder,
    Function,
    Home,
    CatalogElement,
]

ContainerObject = Union[
    Source,
    Dataset,
    Space,
    Folder,
    # Function,
    Home,
]

MAP_RESP = {
    "home": Home,
    "space": Space,
    "dataset": Dataset,
    "folder": Folder,
    "function": Function,
    "source": Source,
    "dataset": Dataset,
    "file": File,
    "tags": Tags,
    "wiki": Wiki,
}

## INPUTS ##

NewCatalogObject = Union[NewFolder, NewDataset, NewSpace, NewSource]

MAP_INPUT = {
    "folder": NewFolder,
    "dataset": NewDataset,
    "space": NewSpace,
    "source": NewSource,
}



from uuid import UUID
from pyarrow import Table
import pytest
from dremio.models.jobs import Job, JobResult
from tests.testutils.helper import random_name
from src.dremio import Dremio, JobResult, Dataset
from pyarrow.flight import FlightStreamReader


def test_query(dremio: Dremio, sample_dataset_paths: list[str]):
    sql = f"SELECT * FROM {sample_dataset_paths[0]}"
    result_arrow = dremio.query(sql, "arrow")
    assert isinstance(result_arrow, JobResult)

    result_http = dremio.query(sql, "http")
    assert isinstance(result_http, JobResult)

    assert result_arrow.rowCount == result_http.rowCount


def test__flight_query_stream(dremio: Dremio, sample_dataset_paths: list[str]):
    sql = f"SELECT * FROM {sample_dataset_paths[0]}"
    stream = dremio._flight_query_stream(sql)
    assert isinstance(stream, FlightStreamReader)


def test_flight_query_stream(dremio: Dremio, sample_dataset_paths: list[str]):
    sql = f"SELECT * FROM {sample_dataset_paths[0]}"
    stream = dremio.flight.query_stream(sql)
    assert isinstance(stream, FlightStreamReader)


def test__flight_query(dremio: Dremio, sample_dataset_paths: list[str]):
    sql = f"SELECT * FROM {sample_dataset_paths[0]}"
    table = dremio._flight_query(sql)
    assert isinstance(table, Table)


def test_flight_query(dremio: Dremio, sample_dataset_paths: list[str]):
    sql = f"SELECT * FROM {sample_dataset_paths[0]}"
    table = dremio.flight.query(sql)
    assert isinstance(table, Table)


def test_flight_query_dataset(dremio: Dremio, dataset: Dataset):
    table = dremio._flight_query_dataset(dataset)
    assert isinstance(table, Table)


def test__flight_query_dataset(dremio: Dremio, dataset: Dataset):
    table = dremio.flight.query_dataset(dataset)
    assert isinstance(table, Table)


def test_flight_query_equal(dremio: Dremio, dataset: Dataset):
    table_sql = dremio.flight.query(dataset.sql or "")
    table_ds = dremio.flight.query_dataset(dataset)

    assert table_ds.schema == table_sql.schema
    assert table_ds.num_rows == table_ds.num_rows


def test_http_query(dremio: Dremio, sample_dataset_paths: list[str]):
    sql = f"SELECT * FROM {sample_dataset_paths[0]}"
    result, job = dremio.http.query(sql)
    assert job.jobState == "COMPLETED"
    assert isinstance(result, JobResult)


def test_http_query_dataset(dremio: Dremio, dataset: Dataset):
    result, job = dremio.http.query_dataset(dataset)
    assert job.jobState == "COMPLETED"
    assert isinstance(result, JobResult)



__all__ = ["_MixinUser"]  # this is like `export ...` in typescript
from typing import Any
import requests
from dataclasses import asdict

from ..utils.converter import clear_at
from ..models import *

from . import BaseClass
from ._catalog import _MixinCatalog


class _MixinUser(_MixinCatalog, BaseClass):

    def get_users(self) -> list[User]:
        """Get the users.

        Returns:
          list[User]: The users.
        """
        url = f"{self.hostname}/api/v3/user"
        response = requests.get(url, headers=self._headers)
        self._raise_error(response)
        return [cast(User, clear_at(user)) for user in response.json()]

    def get_user(self, id: Union[UUID, str]) -> User:
        """Get the user.

        Parameters:
          id: user uuid.

        Returns:
          User: user.
        """
        url = f"{self.hostname}/api/v3/user/{id}"
        response = requests.get(url, headers=self._headers)
        self._raise_error(response)
        return cast(User, clear_at(response.json()))

    def get_user_by_name(self, name: str) -> User:
        """Get the user by name.

        Parameters:
          name: users name.

        Returns:
          User: user.
        """
        url = f"{self.hostname}/api/v3/user/by-name/{name}"
        response = requests.get(url, headers=self._headers)
        self._raise_error(response)
        return cast(User, clear_at(response.json()))

    def create_user(self, user: User) -> User:
        """Create the user.

        Parameters:
          user: new user object.

        Returns:
          User: user object from dremio after creation.
        """
        url = f"{self.hostname}/api/v3/user"
        payload: dict[str, Any] = asdict(user, dict_factory=User.dict_factory)
        response = requests.post(url, headers=self._headers, json=payload)
        self._raise_error(response)
        return cast(User, clear_at(response.json()))

    def update_user(self, id: Union[UUID, str], user: User) -> User:
        """Update the user.

        Parameters:
          id: user uuid.
          user: user object with changes.

        Returns:
          User: user from dremio after changes.
        """
        url = f"{self.hostname}/api/v3/user/{id}"
        payload: dict[str, Any] = asdict(user, dict_factory=User.dict_factory)
        response = requests.put(url, headers=self._headers, json=payload)
        self._raise_error(response)
        return cast(User, clear_at(response.json()))

    def delete_user(self, id: Union[UUID, str], tag: str) -> bool:
        """Delete the user.

          Parameters:
            id: user uuid.
          tag: version tag of current user version.

        Returns:
          bool: True if successful, False otherwise.
        """
        url = f"{self.hostname}/api/v3/user/{id}?version={tag}"
        response = requests.delete(url, headers=self._headers)
        self._raise_error(response)
        if response.status_code not in [200, 204]:
            raise DremioError("unknown error", "", response.status_code)
        return True



from threading import Thread


class ThreadWithReturnValue(Thread):
    def __init__(
        self, group=None, target=None, name=None, args=(), kwargs={}, Verbose=None
    ):
        Thread.__init__(self, group, target, name, args, kwargs)
        self._return = None

    def run(self):
        if self._target is not None:  # type: ignore
            self._return = self._target(*self._args, **self._kwargs)  # type: ignore

    def join(self, *args):
        Thread.join(self, *args)
        return self._return



from dataclasses import dataclass, fields
from typing import Literal, Optional, Any, TypeVar, Union


@dataclass
class Message:
    message: str


T = TypeVar("T")


def cast(T: type[T], d: Any) -> T:
    if isinstance(d, list):  # if already correct type
        d = [cast(T.__args__[0], i) for i in d]  # type: ignore
    try:  # try to cast Optional, Union and Literal
        d = cast(T.__args__[0], d)  # type: ignore
    except:
        pass
    # try to cast dataclass
    if isinstance(d, dict):  # if already correct type
        d_ = {}
        for field in fields(T):  # type: ignore
            try:
                d_[field.name] = cast(field.type, d[field.name])  # type: ignore
            except:
                pass
        d = T(**d_)
    return d



from dataclasses import dataclass
from uuid import UUID
from typing import Optional, Literal, Union
from typing_extensions import Self

from ..exceptions import DremioConnectorError

from .custom_objects import CatalogElement

from .accesscontrol import (
    Owner,
    AccessControlList,
    Privilege,
)


from .baseclasses import DremioAccessible, DremioObject
from ..utils.converter import path_to_dotted


@dataclass
class Folder(DremioAccessible, DremioObject):
    entityType: Literal["folder"]
    id: UUID
    path: list[str]
    tag: str
    children: Optional[list[CatalogElement]] = None
    accessControlList: Optional[AccessControlList] = None
    permissions: Optional[list[Privilege]] = None
    owner: Optional[Owner] = None

    def __iter__(self):
        if not self.children:
            return iter([])
        return iter(self.children)

    def __next__(self):
        if not self.children:
            return iter([])
        for child in self.children:
            yield child

    def __getitem__(self, key: Union[int, slice]) -> list[CatalogElement]:
        if not self.children:
            raise IndexError("This folder has no children")
        if isinstance(key, slice):
            return self.children[key.start : key.stop : key.step]
        else:
            raise IndexError("Key must be slice. For example folder[1:3]")

    def __str__(self) -> str:
        return f"Folder(path = '{path_to_dotted(self.path)}', id = {self.id})"

    def __repr__(self) -> str:
        return f"Folder(path = '{path_to_dotted(self.path)}', id = {self.id})"

    @property
    def name(self) -> str:
        return self.path[-1]

    def pull(self) -> Self:
        """Get the newest version of this folder form dremio.
        Could be useful for multi-threading and updated children.
        This is the opposite of .commit()

        Raises:
            TypeError: raises if no dremio instance is connected

        Returns:
            Self: gives you itself for chaining
        """
        if not self._dremio:
            raise TypeError(
                "This object has no dremio instance!\nTry to commit the dataset directly in the dremio sdk"
            )
        self = self._dremio.get_folder(self.id)
        return self

    def copy(
        self,
        target_path: Union[list[str], str],
        *,
        assume_privileges: bool = True,
        relative_references: bool = False,
        overwrite_existing: bool = False,
    ) -> "Folder":
        """Copy the folder to the target path. Same as `Dremio.copy_folder(this_folder_path, target_path)`. [learn more](https://github.com/continental/pydremio/blob/master/docs/FOLDER.md#copy-a-folder)"""
        if not self._dremio:
            raise TypeError(
                "This Folder has no dremio instance!\nTry to copy the folder directly in the dremio sdk"
            )
        return self._dremio.copy_folder(
            self.path,
            target_path,
            assume_privileges=assume_privileges,
            relative_references=relative_references,
            overwrite_existing=overwrite_existing,
        )

    def reference(
        self,
        target_path: Union[list[str], str],
        *,
        assume_privileges: bool = True,
        overwrite_existing: bool = False,
    ) -> "Folder":
        """Reference the folder to the target path. Same as `Dremio.reference_folder(this_folder_path, target_path)`. [learn more](https://github.com/continental/pydremio/blob/master/docs/FOLDER.md#reference-a-folder)"""
        if not self._dremio:
            raise TypeError(
                "This Folder has no dremio instance!\nTry to reference the folder directly in the dremio sdk"
            )
        return self._dremio.reference_folder(
            self.path,
            target_path,
            assume_privileges=assume_privileges,
            overwrite_existing=overwrite_existing,
        )

    def delete(self, recursive=False) -> None:
        """Delete the folder. Same as `Dremio.delete_folder(this_folder_path)`."""
        if not self._dremio:
            raise TypeError(
                "This Folder has no dremio instance!\nTry to delete the folder directly in the dremio sdk"
            )
        success: bool = False
        try:
            success = self._dremio.delete_folder(self.path, recursive=recursive)
        except Exception as e:
            raise e
        if not success:
            raise DremioConnectorError(
                "Folder not deleted", f"Folder {self.path} could not be deleted"
            )
        del self



from dataclasses import dataclass, asdict
from uuid import UUID
from typing import Any, Type, Union


def to_dict(d) -> dict:
    # print('obj',obj)
    d = asdict(d)
    result = {}
    for key, value in d.items():
        if isinstance(value, UUID):
            result[key] = str(value)
        if isinstance(key, str):
            if key[0] == "_" and key[1] != "_":
                continue
        result[key] = value
    return result


def path_to_list(path: Union[str, list[str]]) -> list[str]:
    if type(path) == str:
        path = (
            path.replace(", ", ".").replace("[", "").replace("]", "").replace("'", '"')
        )
    if type(path) == str and "." in path:
        path = path.split(".")

    # if type(path) == str and '/' in path:
    #   path = path.split('/')

    if type(path) == str:
        path = [path]

    if type(path) == list or isinstance(path, list):
        path = [path.replace('"', "").strip() for path in path]

    return [el for el in list(path) if el]


def path_to_dotted(path: Union[list[str], str]) -> str:
    path = path_to_list(path)
    return '"' + '"."'.join(path) + '"'


def clear_at(d: dict) -> dict:
    res = {}

    for k, v in d.items():
        if k[0] == "@":
            res[k[1:]] = v
            continue
        res[k] = v

    return res



from dataclasses import dataclass
from typing_extensions import Self
from testcontainers.core.container import DockerContainer
from testcontainers.core.waiting_utils import wait_for_logs
from src.dremio import Dremio, FlightConfig


@dataclass(kw_only=True)
class InitUser:
    userName: str = "admin"
    firstName: str = "admin"
    lastName: str = "admin"
    email: str = "admin@admin.de"
    password: str = "admin_password_123"


class DremioContainer(DockerContainer):
    image = "dremio/dremio-oss"

    def __init__(
        self,
        init_user: InitUser = InitUser(),
        bind_ports: dict[int, int | None] = {
            9047: None,
            31010: None,
            32010: None,
            45678: None,
        },
        docker_client_kw: dict | None = None,
        **kwargs,
    ):
        super().__init__(self.image, docker_client_kw, **kwargs)
        for container_port, host_port in bind_ports.items():
            self.with_bind_ports(container_port, host_port)
        self.init_user: InitUser = init_user

    def start(self, auto_user_creation: bool = True) -> Self:
        self = super().start()
        delay = wait_for_logs(self, r"^.*Dremio Daemon Started as master.*$")
        ip = self.get_container_host_ip()
        port = self.get_exposed_port(9047)
        print(f"Container is ready after {delay} seconds")
        print(f"Connect to Dremio at {ip}:{port}")
        if auto_user_creation:
            self.create_firstuser()
        return self

    def create_firstuser(self, user: InitUser | None = None) -> InitUser:
        if not user:
            user = self.init_user
        self.exec(
            f"""\
              curl 'http://localhost:9047/apiv2/bootstrap/firstuser' -X PUT \
              -H 'Authorization: _dremionull' -H 'Content-Type: application/json' \
              --data-binary '{{"userName":"{user.userName}","firstName":"{user.firstName}","lastName":"{user.lastName}","email":"{user.email}","createdAt":1526186430755,"password":"{user.password}"}}'\
            """
        )
        print(f"InitUser {user.userName} created")
        return user

    def add_sample_sources(self, *sample_files_to_format: str) -> list[str]:
        "returns a list of paths for the created sample datasets"
        print("import samples")
        user = self.dremio.login(self.init_user.userName, self.init_user.password)
        res = self.exec(
            f"""
              curl 'http://localhost:9047/apiv2/source/Samples' -X PUT \
              -H 'Authorization: _dremio{user.token}' -H 'Content-Type: application/json' \
              --data-binary '{{"config":{{"externalBucketList":["samples.dremio.com"],"credentialType":"NONE","secure":false,"propertyList":[]}},"name":"Samples","accelerationRefreshPeriod":3600000,"accelerationGracePeriod":10800000,"accelerationNeverRefresh":true,"accelerationNeverExpire":true,"accelerationActivePolicyType":"PERIOD","accelerationRefreshSchedule":"0 0 8 * * *","type":"S3"}}'
            """
        )
        for sample in sample_files_to_format:
            print("formating", sample)
            res = self.exec(
                f"""
                  curl 'http://localhost:9047/apiv2/source/Samples/file_format/samples.dremio.com/{sample}' -X PUT \
                  -H 'Authorization: _dremio{user.token}' -H 'Content-Type: application/json' \
                  --data-binary '{{"fieldDelimiter":",","quote":"\\"","comment":"#","lineDelimiter":"\\r\\n","escape":"\\"","extractHeader":true,"trimHeader":true,"skipFirstLine":false,"type":"Text"}}'
                """
            )
        return [
            f'Samples."samples.dremio.com"."{sample}"'
            for sample in sample_files_to_format
        ]

    @property
    def dremio(self) -> Dremio:
        hostname = self.get_container_host_ip()
        port = self.get_exposed_port(9047)
        grpc_port = self.get_exposed_port(32010)
        return Dremio(
            hostname=hostname,
            port=int(port),
            protocol="http",
            flight_config=FlightConfig(port=int(grpc_port)),
            username=self.init_user.userName,
            password=self.init_user.password,
        )

    def __enter__(self) -> tuple[Self, Dremio]:
        return self, self.dremio


if __name__ == "__main__":
    from time import sleep

    with DremioContainer() as container:
        sleep(300)



__all__ = ["_MixinFlight"]  # this is like `export ...` in typescript
import copy
import logging
import pyarrow
from pyarrow.flight import (
    FlightClient,
    FlightStreamReader,
    FlightCallOptions,
    FlightDescriptor,
    FlightUnavailableError,
    FlightInfo,
)

from ..models import *

from . import BaseClass
from ._dataset import _MixinDataset


class _MixinFlight(_MixinDataset, BaseClass):

    @property
    def flight_url(self):
        return self.flight_config.uri(self.hostname)

    def _flight_client(
        self, flight_config: Optional[FlightConfig] = None
    ) -> FlightClient:
        """Get the pre-configured pyarrow.flight.FlightClient with optional configuration.

        Args:
            flight_config (Optional[FlightConfig], optional): Overwrite pre-config. Defaults to None.

        Returns:
            FlightClient: [pyarrow.flight.FlightClient](https://arrow.apache.org/docs/python/generated/pyarrow.flight.FlightClient.html)
        """
        flight_config = flight_config or self.flight_config
        return FlightClient(
            location=(flight_config.uri(self.hostname)),
            disable_server_verification=flight_config.disable_certificate_verification,
            tls_root_certs=flight_config.tls_root_certs,  # type: ignore
        )

    def _flight_options(
        self, flight_config: Optional[FlightConfig] = None
    ) -> FlightCallOptions:
        flight_config = flight_config or self.flight_config
        return FlightCallOptions(
            headers=flight_config.get_headers(
                {"authorization": f"bearer {self._token}"},
                as_bytes=True,
            )  # type: ignore
        )

    def _flight_get_flight_info(
        self, descriptor: FlightDescriptor, options: FlightCallOptions | None = None
    ) -> FlightInfo:
        """Wrapper around pyarrow.flight.get_flight_info with pre-config.

        Args:
            descriptor (FlightDescriptor)
            options (FlightCallOptions | None, optional)

        Returns:
            FlightInfo
        """
        options = options or self._flight_options()
        return self._flight_client().get_flight_info(descriptor, options)

    def _flight_query_stream(
        self,
        sql_request: Union[str, SQLRequest],
        *,
        flight_config: Optional[FlightConfig] = None,
        flight_options: Optional[FlightCallOptions] = None,
    ) -> FlightStreamReader:
        """Execute a SQL request and get the results as flight data. [learn more](https://arrow.apache.org/docs/python/generated/pyarrow.Table.html)
        You can use the `.to_pandas()` method on the result to convert it to a pandas DataFrame.
        Use Dataset.run() for easier handling.

        Parameters:
          sql_request (str|SQLRequest): SQL request as SQLRequest object.
          flight_config (FlightConfig | None): Optional: Config for flight. Leave empty to use `Dremio.flight_config`.

        Returns:
          pyarrow.flight.FlightStreamReade: The job results as [pyarrow.flight.FlightStreamReade](https://arrow.apache.org/docs/python/generated/pyarrow.flight.FlightStreamReader.htm).
        """
        if isinstance(sql_request, str):
            sql_request = SQLRequest(sql_request)
        flight_config = flight_config or self.flight_config

        client = self._flight_client(flight_config)
        options = flight_options or self._flight_options(flight_config)

        try:
            flight_info = client.get_flight_info(
                FlightDescriptor.for_command(sql_request.sql), options
            )
        except FlightUnavailableError as e:
            if not flight_config.allow_autoconfig:
                raise e
            if "SETTINGS" in str(e) and flight_config.tls == False:
                logging.warning(
                    "Got FlightUnavailableError. Retrying with `flight_config.tls = True`..."
                )
                flight_config_copy = copy.copy(flight_config)
                flight_config_copy.tls = True
                results = self._flight_query_stream(
                    sql_request=sql_request, flight_config=flight_config_copy
                )
                logging.warning(
                    "Retry with `flight_config.tls = True` was successful. Set `Dremio.flight_config.tls = True` for future queries."
                )
                return results
            if (
                "empty address list" in str(e)
                and self.flight_config.disable_certificate_verification == False
            ):
                logging.warning(
                    "Got FlightUnavailableError. Retry with `flight_config.disable_certificate_verification = True`..."
                )
                flight_config_copy = copy.copy(flight_config)
                flight_config_copy.disable_certificate_verification = True
                results = self._flight_query_stream(
                    sql_request=sql_request, flight_config=flight_config_copy
                )
                logging.warning(
                    "Retry with `flight_config.disable_certificate_verification = True` was successful. Set `Dremio.flight_config.disable_certificate_verification = True` for future queries."
                )
                return results
            raise e
        return client.do_get(flight_info.endpoints[0].ticket, options)

    def _flight_query(
        self,
        sql_request: Union[str, SQLRequest],
        *,
        flight_config: Optional[FlightConfig] = None,
        flight_options: Optional[FlightCallOptions] = None,
    ) -> pyarrow.Table:
        """Execute a SQL request and get the results as flight data. [learn more](https://arrow.apache.org/docs/python/generated/pyarrow.Table.html)
        You can use the `.to_pandas()` method on the result to convert it to a pandas DataFrame.
        Use Dataset.run() for easier handling.

        Parameters:
          sql_request (str|SQLRequest): SQL request as SQLRequest object.
          flight_config (FlightConfig | None): Optional: Config for flight. Leave empty to use `Dremio.flight_config`.

        Returns:
          pyarrow.Table: The job results as pyarrow.Table.
        """
        return self._flight_query_stream(
            sql_request=sql_request,
            flight_config=flight_config,
            flight_options=flight_options,
        ).read_all()

    def _flight_query_dataset(
        self,
        dataset: Dataset | list[str] | str | None = None,
        *,
        id: Union[UUID, str, None] = None,
        flight_config: Optional[FlightConfig] = None,
        flight_options: Optional[FlightCallOptions] = None,
    ) -> pyarrow.Table:
        """Get the dataset as flight data. [learn more](https://arrow.apache.org/docs/python/generated/pyarrow.Table.html)
        You can use the `.to_pandas()` method on the result to convert it to a pandas DataFrame.
        Use Dataset.run() for easier handling.

        Parameters:
          dataset: Dataset | path: Instance of a Dataset or "A.B.C" or ["A","B","C"].
          id: dataset uuid.

        Returns:
          pyarrow.Table: The job results as pyarrow.Table.
        """
        if not dataset:
            if not id:
                raise TypeError("A query on a dataset needs path or id")
            dataset = self.get_dataset(id=id)
        if isinstance(dataset, list) or isinstance(dataset, str):
            dataset = self.get_dataset(path=dataset)
        if not isinstance(dataset, Dataset):
            raise TypeError("Dataset is needed to run a dataset")

        if not dataset.sql:
            raise DremioConnectorError(
                "No SQL statement in dataset",
                f"Please check the sql statement in dataset {dataset.id}",
            )
        return self._flight_query(
            dataset.sql,
            flight_config=flight_config,
            flight_options=flight_options,
        )

    @property
    def flight(self):
        """Namespace: Collection of methods related to arrow flight.

        Returns:
            Flight: Namespace of arrow flight related methods
        """
        return Flight(self)


class Flight:
    """Namespace: Collection of methods related to arrow flight."""

    def __init__(self, mixin: _MixinFlight):
        self.query = mixin._flight_query
        self.query_stream = mixin._flight_query_stream
        self.query_dataset = mixin._flight_query_dataset
        self.config = mixin.flight_config
        self.url = mixin.flight_url
        self.get_client = mixin._flight_client
        self.get_options = mixin._flight_options
        self.get_flight_info = mixin._flight_get_flight_info

    @property
    def client(self) -> FlightClient:
        """Shortcut for get_client() with default config

        Returns:
            FlightClient: [pyarrow.flight.FlightClient](https://arrow.apache.org/docs/python/generated/pyarrow.flight.FlightClient.html)
        """
        return self.get_client()

    @property
    def options(self) -> FlightCallOptions:
        """Shortcut for get_options() with default config."""
        return self.get_options()



import pytest

from tests.testutils.helper import random_name
from src.dremio import Dremio, DremioError, Space


def test_create_get_delete_a_folder(dremio: Dremio, space: Space):
    path = [space.name, random_name()]
    folder = dremio.create_folder(path)
    assert folder.path == path
    assert folder.name == path[-1]
    folder.delete()
    with pytest.raises(DremioError):
        dremio.get_folder(path)


def test_create_and_delete_a_subfolder(dremio: Dremio, space: Space):
    parent_path = [space.name, random_name()]
    path = parent_path + [random_name()]
    subfolder = dremio.create_folder(path)
    assert subfolder.path == path
    assert subfolder.name == path[-1]
    subfolder.delete()
    with pytest.raises(
        DremioError
    ):  # should not find this folder, cz it should been deleted
        dremio.get_folder(path)
    parent_folder = dremio.get_folder(parent_path)
    assert parent_folder.name == parent_path[-1]
    parent_folder.delete()
    with pytest.raises(
        DremioError
    ):  # should not find this folder, cz it should been deleted
        dremio.get_folder(parent_path)



__all__ = ["_MixinQuery"]  # this is like `export ...` in typescript
import logging
from typing import overload
from polars import sql
import pyarrow as pa
from pyarrow import flight
from pyarrow.flight import FlightClient

from ..models import *

from . import BaseClass
from ._dataset import _MixinDataset
from ._sql import _MixinSQL
from ._flight import _MixinFlight


class _MixinQuery(_MixinFlight, _MixinDataset, _MixinSQL, BaseClass):
    "This is a abstraction of arrow flight and sql for easer usability."

    @overload
    def query(
        self,
        sql_request: Union[str, SQLRequest],
        method: Literal["arrow"] = "arrow",
    ) -> JobResult:
        """Query Dremio data with SQL.

        Args:
            sql_request (Union[str, SQLRequest]): SQL statement as string or Dremio SQL-object.
            method (Literal[&quot;arrow&quot;, &quot;http&quot;], optional): Query method. Defaults to "arrow".

        Raises:
            DremioConnectorError: Something went wrong while job execution.
            Exception: Any error from arrow flight.

        Returns:
            JobResult
        """

    @overload
    def query(
        self,
        sql_request: Union[str, SQLRequest],
        method: Literal["http"] = "http",
        *,
        limit: Union[int, None] = None,
        offset: int = 0,
        timeout_in_sec: int = 1000,
    ) -> JobResult:
        """Execute a SQL request and get the results. This method don't raise errors.
        Overloaded!!!

        Args:
            sql_request: SQL request as SQLRequest object.
            limit: The limit of the results rows for pagination.
            offset: The offset of the results rows pagination.
            timeout_in_sec: seconds until TimeoutError raising
            timeout_in_sec (int, optional): _description_. Defaults to 1000.

        Returns:
          tuple[JobResult|None, Job]: The job results and the job.
        """

    def query(
        self,
        sql_request: Union[str, SQLRequest],
        method: Literal["arrow", "http"] = "arrow",
        *,
        limit: Union[int, None] = None,
        offset: int = 0,
        timeout_in_sec: int = 1000,
    ) -> JobResult:
        if method == "http":
            job = self._http_start_job(
                sql_request=sql_request,
            )
            if not job.id:
                raise DremioConnectorError(
                    "Error while starting job on sql query",
                    f'Failed to start job on "{sql_request}"',
                )
            return self._http_wait_for_job_result(
                id=job.id,
                limit=limit,
                offset=offset,
                timeout_in_sec=timeout_in_sec,
            )
        elif method == "arrow":
            try:
                table: pa.Table = self._flight_query(sql_request=sql_request)
                return JobResult.from_arrow_table(table)
            except Exception as e:
                logging.warning(
                    "Query over arrow flight failed. Check your config or try again over http."
                )
                raise e
        else:
            raise ValueError('method must be "arrow" or "http"')



__all__ = ["Dremio"]  # this is like `export ...` in typescript

import logging
from urllib import parse
import requests
from requests.exceptions import JSONDecodeError


from ..utils.validation import Validation
from ..utils.converter import path_to_list
from ..models import *


class Dremio:

    v = Validation()

    def __init__(
        self,
        hostname: str,
        *,
        username: Union[str, None] = None,
        password: Union[str, None] = None,
        auth: Union[str, None] = None,
        token: Union[str, None] = None,
        modus: Literal["stand_alone", "server"] = "stand_alone",
        port: Union[int, None] = None,
        protocol: Union[Literal["https", "http"], None] = None,
        flight_config: FlightConfig | FlightConfigDict = FlightConfig(),
        loglevel: (
            Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", "NOTSET"] | None
        ) = "INFO",
    ):
        """
        Create a new Dremio instance.
        Its possible to set the authentification directly or later with the `auth` or `login` method.

        Parameters:
          hostname: hostname of the Dremio server (protocol and port optional).
          username: dremio username, e-mail.
          password: password or token from dremio ui.
          auth: authentification: _dremio<token>.
          token: token from POST /login on dremio api.
          modus: stand_alone or server, default: stand_alone
          port: port of the Dremio server, if None it will be auto set.
          protocol: https/http protocol of the Dremio server.
          loglevel: 'DEBUG'|'INFO'|'WARNING'|'ERROR'|'CRITICAL'|'NOTSET'|None to set the log level. Set None to prevent any setting.

        Returns:
          Dremio: Dremio instance.
        """
        v = self.v
        self.stand_alone = modus == "stand_alone"
        self.hostname: str = v.hostname(hostname, port, protocol)
        self.flight_config = (
            FlightConfig(**flight_config)
            if isinstance(flight_config, dict)
            else flight_config
        )
        self._headers: dict[str, str] = {"Content-Type": "application/json"}
        if loglevel:
            self.set_log_level(loglevel)
        if auth or token:
            if token:
                auth = f"_dremio{token}"
                self._token = token
            if not auth:
                raise AttributeError("")
            self._token = auth[7:]
            self._headers["Authorization"] = auth
            return
        if username and password:
            self.login(username, password)
            return

    def set_log_level(
        self,
        loglevel: (
            Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL", "NOTSET"] | None
        ) = None,
    ) -> int | None:
        """Set log level.

        Args:
            loglevel (Literal[&#39;DEBUG&#39;,&#39;INFO&#39;, &#39;WARNING&#39;, &#39;ERROR&#39;, &#39;CRITICAL&#39;, &#39;NOTSET&#39;] | None, optional): _description_. Defaults to None.

        Returns:
            int|None: logging.<LOGLEVEL> or None
        """
        if not loglevel:
            return None
        elif loglevel == "DEBUG":
            level = logging.DEBUG
        elif loglevel == "INFO":
            level = logging.INFO
        elif loglevel == "WARNING":
            level = logging.WARNING
        elif loglevel == "ERROR":
            level = logging.ERROR
        elif loglevel == "CRITICAL":
            level = logging.CRITICAL
        elif loglevel == "NOTSET":
            level = logging.NOTSET
        logging.basicConfig(level=level)
        return level

    # @staticmethod
    # def from_env(
    #     hostname: str = "DREMIO_HOSTNAME",
    #     username: str = "DREMIO_USERNAME",
    #     password: str = "DREMIO_PASSWORD",
    #     modus: str = "DREMIO_MODUS",
    #     port: str = "DREMIO_PORT",
    #     protocol: str = "DREMIO_PROTOCOL",
    #     *,
    #     flight_port: str = "DREMIO_FLIGHT_PORT",
    #     flight_tls: str = "DREMIO_FLIGHT_TLS",
    #     flight_path_to_certs: str = "DREMIO_FLIGHT_PATH_TO_CERTS",
    #     load_dotenv: bool = False,
    # ) -> "Dremio":
    #     """Create a new Dremio instance from environment variables. [learn more](https://github.com/continental/pydremio/blob/master/docs/DREMIO_LOGIN.md)
    #     The environment variables are:
    #       - DREMIO_HOSTNAME
    #       - DREMIO_USERNAME
    #       - DREMIO_PASSWORD
    #       - DREMIO_MODUS
    #       - DREMIO_PORT
    #       - DREMIO_PROTOCOL
    #       - DREMIO_FLIGHT_PORT
    #       - DREMIO_FLIGHT_TLS
    #       - DREMIO_FLIGHT_PATH_TO_CERTS


    #     Parameters:
    #       hostname: environment variable name for the hostname.
    #       username: environment variable name for the username.
    #       password: environment variable name for the password.
    #       modus: environment variable name for the modus.
    #       port: environment variable name for the port.
    #       protocol: environment variable name for the protocol.
    #       load_dotenv: load the dotenv file. Default: False.

    #     Returns:
    #       Dremio: Dremio instance.
    #     """
    #     if load_dotenv:
    #         from dotenv import load_dotenv as ldotenv

    #         ldotenv()
    #     import os

    #     hostname = os.getenv(hostname) or os.getenv("DREMIO_ENDPOINT")  # type: ignore
    #     username = os.getenv(username)  # type: ignore
    #     password = os.getenv(password)  # type: ignore
    #     modus = os.getenv(modus) or "stand_alone"
    #     port = os.getenv(port) or "443"
    #     protocol = os.getenv(protocol) or "https"
    #     if not hostname:
    #         raise ValueError("No hostname provided")
    #     if not username or not password:
    #         logging.warning("No username or password provided")
    #     if modus not in ["stand_alone", "server"]:
    #         raise ValueError("Modus must be either stand_alone or server")
    #     try:
    #         int(port)
    #     except:
    #         raise ValueError("Port must be a valid integer")
    #     if protocol not in ["https", "http"]:
    #         raise ValueError("Protocol must be either https or http")

    #     flight_conf: FlightConfigDict = {}
    #     if f_port := os.getenv(flight_port):
    #         flight_conf["port"] = int(f_port)
    #     if f_tls := os.getenv(flight_tls):
    #         flight_conf["tls"] = bool(f_tls)
    #     if f_path_to_certs := os.getenv(flight_path_to_certs):
    #         flight_conf["path_to_certs"] = str(f_path_to_certs)

    #     return Dremio(hostname, username=username, password=password, modus=modus, port=int(port), protocol=protocol, flight_config=flight_conf)  # type: ignore

    def auth(self, auth: Union[str, None] = None, token: Union[str, None] = None):
        """Set the authentification.
        Parameters:
          auth: authentification: _dremio<token>.
          token: token from POST /login on dremio api.
        Returns:
          Dremio: new logged in dremio instance.
        """
        modus: Literal["stand_alone", "server"] = (
            "stand_alone" if self.stand_alone else "server"
        )
        return Dremio(hostname=self.hostname, auth=auth, token=token, modus=modus)

    def login(self, username: str, password: str) -> CurrentUser:
        """
        Log in to Dremio. Its overwitring the current authentification.

        Parameters:
          username: dremio username, e-mail.
          password: password or token from dremio ui.

        Returns:
          CurrentUser: The current user.
        """
        v = self.v
        self.username = v(username=username)
        url = f"{self.hostname}/apiv2/login"
        payload = {"userName": username, "password": password}
        response = requests.post(url, json=payload)
        self._raise_error(response)
        user = CurrentUser(**response.json())
        self.username = user.userName
        self._token = user.token
        self._headers["Authorization"] = f"_dremio{user.token}"
        return user

    def _T(self, obj: dict, input: bool = False) -> CatalogObject:
        containerType = obj["entityType"]
        if input:
            T: Type = MAP_INPUT[containerType]
        else:
            # print(obj)
            T: Type = MAP_RESP[containerType]
        try:
            typed_object = cast(T, obj)
        except TypeError as e:
            raise TypeError(f"Error typing {obj} as {T}. {e}")
        return typed_object

    def _stand_alone_404(self, response: Response) -> bool:
        """This privat method is for stand alone use:
        If a resource is not found in stand alone mode, it should be returned None.
        On a server (not stand alone) 404 should be returned to the client.

        Args:
            response (Response): requests.Response

        Returns:
            bool: True, if in stand alone mode and response is 404
        """
        return self.stand_alone and response.status_code == 404

    def _raise_error(self, response: Response):
        status = response.status_code
        if status == 401:
            raise AuthError("Login incorrect or user not found")
        if status not in [200, 204]:
            try:
                raise DremioError(status_code=status, **response.json())
            except JSONDecodeError:
                # only if can not parse the Dremio Error
                raise DremioError(status_code=status, errorMessage="", moreInfo="")

    def _get_url_of_object(
        self,
        path: Union[str, list[str], None] = None,
        *,
        id: Union[UUID, str, None] = None,
    ) -> str:
        """Generate an url for any given Dremio Object.

        Args:
            path (Union[str,list[str],None], optional): path to Dremio Object
            id (Union[UUID,str,None], optional): id of Dremio Object

        Returns:
            str: url
        """
        obj = self._get_catalog_object(id, path)
        subpath = "/".join(path_to_list(obj["path"][:-1]))
        if obj["entityType"] == "folder":
            subpath += "/"
        if obj["entityType"] == "dataset":
            subpath += "."
        subpath += obj["path"][-1]
        return f"{self.hostname}/space/{subpath}"

    def _get_catalog_object(
        self,
        id: Union[UUID, str, None] = None,
        path: Union[str, list[str], None] = None,
    ) -> dict:  # CatalogObject|list[CatalogObject]:
        if not path and not id:
            raise ValueError("No path or id provided")
        url = f"{self.hostname}/api/v3/catalog/{str(id) or ''}"
        if path:
            path = path_to_list(path)
            if isinstance(path, list):
                path = "/".join(path)
            quoted_path = parse.quote(path).replace(" ", "%20").replace("%2F", "/")
            url = f"{self.hostname}/api/v3/catalog/by-path/{quoted_path}"
        response = requests.get(url, headers=self._headers)
        if response.status_code == 401:
            raise DremioError("Unauthorized", "request not allowed", status_code=401)
        self._raise_error(response)
        response.raise_for_status()
        result = response.json()
        if "data" in result:
            return result["data"]
        return result



__all__ = [
    "BaseClass",
    "_MixinCatalog",
    "_MixinDataset",
    "_MixinFolder",
    "_MixinSQL",
    "_MixinFlight",
    "_MixinReflection",
    "_MixinUser",
    "_MixinRole",
    "_MixinQuery",
]

from .baseclass import Dremio as BaseClass
from ._catalog import _MixinCatalog
from ._dataset import _MixinDataset
from ._folder import _MixinFolder
from ._sql import _MixinSQL
from ._flight import _MixinFlight
from ._reflection import _MixinReflection
from ._user import _MixinUser
from ._role import _MixinRole
from ._query import _MixinQuery



__all__ = ["_MixinRole"]  # this is like `export ...` in typescript
from typing import Any
import requests
from ..utils.converter import clear_at
from ..models import *

from . import BaseClass
from ._catalog import _MixinCatalog


class _MixinRole(_MixinCatalog, BaseClass):

    def get_role(self, id: Union[UUID, str]) -> Role:
        """Get the role.

        Parameters:
          id: role uuid.

        Returns:
          Role: role object.
        """
        url = f"{self.hostname}/api/v3/role/{id}"
        response = requests.get(url, headers=self._headers)
        self._raise_error(response)
        return cast(Role, clear_at(response.json()))

    def get_role_by_name(self, name: str) -> Role:
        """Get the role by name.

        Parameters:
          name: role name.

        Returns:
          Role: role object.
        """
        url = f"{self.hostname}/api/v3/role/by-name/{name}"
        response = requests.get(url, headers=self._headers)
        self._raise_error(response)
        return cast(Role, clear_at(response.json()))

    def create_role(
        self,
        name: str,
        roles: Union[list[Role], None] = None,
        description: Union[str, None] = None,
    ) -> Role:
        """Create the role.

        Parameters:
          name: role name.
          roles: list of roles.
          description: role description.

        Returns:
          Role: role object from dremio after creation.
        """
        url = f"{self.hostname}/api/v3/role"
        payload: dict[str, Any] = {"name": name}
        if roles:
            payload["roles"] = [{"id": role.id} for role in roles]
        if description:
            payload["description"] = description
        response = requests.post(url, headers=self._headers, json=payload)
        self._raise_error(response)
        return cast(Role, clear_at(response.json()))

    def delete_role(self, id: Union[str, UUID]) -> bool:
        """Delete the role.

        Parameters:
          role: role object.

        Returns:
          bool: True if successful, raise error otherwise.
        """
        url = f"{self.hostname}/api/v3/role/{id}"
        response = requests.delete(url, headers=self._headers)
        self._raise_error(response)
        if response.status_code not in [200, 204]:
            raise DremioError("unknown error", "", response.status_code)
        return True



from dataclasses import dataclass, field
import json
from uuid import UUID
from typing import Optional, Any, Literal, NewType, Union
from datetime import datetime

from .accesscontrol import Owner, AccessControlList, Privilege, Permissions


@dataclass
class CatalogElement:
    id: UUID
    path: list[str]
    type: Literal["CONTAINER", "DATASET"]  # |
    tag: Optional[str] = None
    createdAt: Optional[datetime] = None
    containerType: Optional[
        Literal["SPACE", "SOURCE", "FOLDER", "HOME", "FUNCTION"]
    ] = None
    datasetType: Optional[Literal["VIRTUAL", "PROMOTED", "DIRECT"]] = None


@dataclass
class DatasetFormat:
    type: Literal["Text", "JSON", "Parquet", "Excel", "XLS", "Delta", "Iceberg"]
    fieldDelimiter: Optional[str] = None
    lineDelimiter: Optional[str] = None
    quote: Optional[str] = None
    comment: Optional[str] = None
    escape: Optional[str] = None
    skipFirstLine: Optional[bool] = None
    extractHeader: Optional[bool] = None
    trimHeader: Optional[bool] = None
    autoGenerateColumnNames: Optional[bool] = None


@dataclass
class Filter:
    condition: Literal["LIKE", "=", "IN"]
    field: str
    value: str
    negation: Optional[bool]


@dataclass
class Filters:
    multiple: bool
    filters: list[Filter]
    operator: Optional[Literal["AND", "OR"]]


@dataclass
class DremioQuery:
    from_condition: str
    select_condition: str
    limit: int
    offset: int
    sort_order: Literal["asc", "desc"]
    where_condition: Optional[Filters]
    jobId: Optional[str]
    sort_property: Optional[str]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Dremio connector to use most of the Dremio-API functionality in python.
All requests and responses are typed as dataclasses. For further information read our README or visit the repo on github.
"""
__author__ = "Holger Zernetsch, Jan Pietsch"
__copyright__ = "Copyright 2024, Continental AG"
__version__ = "0.3.1"



from dataclasses import dataclass
from typing import Optional, TypedDict
from numpy import isin
import pyarrow


class SQLType(TypedDict):
    "This appears, if the JobResult is queried via http"
    name: str  # any sql type


class SchemaField[T: pyarrow.DataType | SQLType](TypedDict):
    "This appears, if the JobResult is queried via arrow flight"
    name: str
    type: T


class Schema[T: pyarrow.DataType | SQLType](list):
    def __init__(self, iterable: list[SchemaField[T]]):
        super().__init__(SchemaField[T](item) for item in iterable)

    def __setitem__(self, index, item):
        super().__setitem__(index, SchemaField[T](item))

    def __getitem__(self, key: int | str):
        for item in self:
            try:
                if item["name"] == key:
                    return item
            except:
                continue
        return super().__getitem__(key)  # type: ignore

    def insert(self, index, item):
        super().insert(index, SchemaField[T](item))

    def append(self, item):
        super().append(SchemaField[T](item))

    def extend(self, other):
        if isinstance(other, type(self)):
            super().extend(other)
        else:
            super().extend(SchemaField[T](item) for item in other)

    @property
    def names(self) -> list[str]:
        return [f["name"] for f in self]

    @property
    def types(self) -> list[T]:
        return [f["type"] for f in self]

    @property
    def type_names(self) -> list[str]:
        types = self.types
        if isinstance(types[0], dict):
            return [t["name"] for t in types]  # type: ignore
        return [str(t) for t in types]



__all__ = ["_MixinReflection"]  # this is like `export ...` in typescript

import logging
from typing import Any
import requests
from time import sleep

from ..utils.converter import to_dict
from ..models import *

from . import BaseClass
from ._catalog import _MixinCatalog


class _MixinReflection(_MixinCatalog, BaseClass):

    def get_reflections(self) -> list[Reflection]:
        """Get all reflections.

        Returns:
          list[Reflection]: reflections.
        """
        url = f"{self.hostname}/api/v3/reflection"
        response = requests.get(url, headers=self._headers)
        self._raise_error(response)
        try:
            data = response.json()["data"]
        except KeyError:
            raise KeyError("Expected data-tag in reflections response, but not found")
        return [cast(Reflection, reflection) for reflection in data]

    def get_reflection(self, reflection_id: Union[UUID, str]) -> Reflection:
        """Get one reflection.

        Parameters:
          id: reflection uuid.

        Returns:
          Reflection: reflection object.
        """
        url = f"{self.hostname}/api/v3/reflection/{reflection_id}"
        response = requests.get(url, headers=self._headers)
        self._raise_error(response)
        return cast(Reflection, response.json())

    def create_reflection(
        self, dataset_id: str | UUID, name: str, reflection: NewReflection
    ) -> Reflection:
        url = f"{self.hostname}/api/v3/reflection"
        payload = to_dict(reflection) | {
            "datasetId": dataset_id,
            "name": name,
        }
        response = requests.post(url, headers=self._headers, json=payload)
        self._raise_error(response)
        return cast(Reflection, response.json())

    def update_reflection(self, reflection: Reflection) -> Reflection:
        """Update the reflection.

        Parameters:
          reflection: reflection object.

        Returns:
          Reflection: updated reflection.
        """
        url = f"{self.hostname}/api/v3/reflection/{reflection.id}"
        payload: dict[str, Any] = to_dict(reflection)
        response = requests.put(url, headers=self._headers, json=payload)
        self._raise_error(response)
        return cast(Reflection, response.json())

    def delete_reflection(self, reflection_id: UUID | str) -> None:
        url = f"{self.hostname}/api/v3/reflection/{reflection_id}"
        response = requests.delete(url, headers=self._headers)
        self._raise_error(response)
        if response.status_code not in [200, 204]:
            raise DremioConnectorError(
                "Deletion of reflection failed",
                f"Failed to delete reflection {reflection_id}",
            )
        return None

    def refresh_reflection(
        self, reflection_id: UUID | str, verbose: bool = False
    ) -> Reflection:
        """Update the reflection.

        Parameters:
          reflection_id: reflection uuid.
          verbose: Write infos and warnings into the default logger. Default: False.

        Returns:
          Reflection: refreshed reflection.
        """
        if verbose:
            logging.info(f"Updating reflection {reflection_id}")
        # load reflection to disable and reenable it
        reflection = self.get_reflection(reflection_id)
        reflection.enabled = False
        refreshing_reflection = self.update_reflection(reflection)
        del reflection
        sleep(1)
        refreshing_reflection.enabled = True
        _ = self.update_reflection(refreshing_reflection)
        del refreshing_reflection
        sleep(1)

        # retrying
        retried = False
        while True:
            refreshed_reflection = self.get_reflection(reflection_id)
            status = refreshed_reflection.status.combinedStatus
            if status == "checkup_reflection":
                if verbose:
                    logging.info(
                        f"SUCCESS: {refreshed_reflection.name} - Reflection was refreshed"
                    )
                break
            elif status == "REFRESHING":
                sleep(1)
                continue
            else:
                if status in {"DISABLED", "INVALID", "CANNOT_ACCELERATE_MANUAL"}:
                    if verbose:
                        logging.warn(
                            f"{refreshed_reflection.name} - Reflection refresh cannot be completed. Reflection status: {status}"
                        )
                elif status in {
                    "FAILED",
                    "INCOMPLETE",
                    "EXPIRED",
                    "CAN_ACCELERATE_WITH_FAILURES",
                    "CANNOT_ACCELERATE_SCHEDULED",
                }:
                    if verbose:
                        logging.warning(
                            (
                                f"{refreshed_reflection.name} - Reflection refresh was not successful. Reflection status: {status}"
                            )
                        )
                    if not retried:
                        sleep(5)
                        if verbose:
                            logging.info("Retrying...")
                        retried = True
                        continue
                break
        return refreshed_reflection



from dataclasses import dataclass
from typing import Literal, Optional, List
from enum import Enum
from datetime import datetime
from uuid import UUID


@dataclass
class DimensionField:
    name: str
    granularity: Optional[str] = None


@dataclass
class Field:
    name: str


@dataclass
class MeasureField:
    name: str
    measureTypeList: List[
        Literal["COUNT", "SUM", "APPROX_COUNT_DISTINCT", "MIN", "MAX", "UNKNOWN"]
    ]


@dataclass
class Status:
    config: Literal["OK", "INVALID"]
    refresh: Literal["GIVEN_UP", "MANUAL", "RUNNING", "SCHEDULED"]
    availability: Literal["NONE", "INCOMPLETE", "EXPIRED", "AVAILABLE"]
    combinedStatus: Literal[
        "CAN_ACCELERATE",
        "CAN_ACCELERATE_WITH_FAILURES",
        "CANNOT_ACCELERATE_MANUAL",
        "CANNOT_ACCELERATE_SCHEDULED",
        "DISABLED",
        "EXPIRED",
        "FAILED",
        "INVALID",
        "INCOMPLETE",
        "REFRESHING",
    ]
    failureCount: int
    lastDataFetch: datetime
    expiresAt: datetime


@dataclass
class Reflection:
    id: UUID
    type: Literal["RAW", "AGGREGATION"]
    name: str
    tag: str
    createdAt: datetime
    updatedAt: datetime
    datasetId: UUID
    currentSizeBytes: int
    totalSizeBytes: int
    enabled: bool
    status: Status
    entityType: Optional[str] = None
    canAlter: Optional[bool] = None
    partitionDistributionStrategy: Optional[str] = None
    sortFields: Optional[List[Field]] = None
    partitionFields: Optional[List[Field]] = None
    distributionFields: Optional[List[Field]] = None
    arrowCachingEnabled: Optional[bool] = None
    canView: Optional[bool] = None
    dimensionFields: Optional[List[DimensionField]] = None
    measureFields: Optional[List[MeasureField]] = None
    displayFields: Optional[List[Field]] = None


@dataclass
class NewReflection:
    type: Literal["RAW", "AGGREGATION"]
    enabled: bool
    entityType: Optional[str] = None
    canAlter: Optional[bool] = None
    partitionDistributionStrategy: Optional[str] = None
    sortFields: Optional[List[Field]] = None
    partitionFields: Optional[List[Field]] = None
    distributionFields: Optional[List[Field]] = None
    arrowCachingEnabled: Optional[bool] = None
    canView: Optional[bool] = None
    dimensionFields: Optional[List[DimensionField]] = None
    measureFields: Optional[List[MeasureField]] = None
    displayFields: Optional[List[Field]] = None



import pytest
from tests.testutils.helper import random_name
from src.dremio import Dremio, DremioError, Folder, Dataset, path_to_dotted


def test_create_delete_dataset(
    dremio: Dremio, folder: Folder, sample_dataset_paths: list[str]
):
    path = folder.path + [random_name()]
    sql = f"SELECT * FROM {sample_dataset_paths[0]}"
    created_dataset = dremio.create_dataset(
        path, sql
    )  # created with testutils.conatiner
    assert created_dataset.path == path
    assert created_dataset.sql == sql
    created_dataset.delete()
    with pytest.raises(DremioError):
        dremio.get_catalog_by_id(created_dataset.id)


def test_get_dataset_by_id(dremio: Dremio, dataset: Dataset):
    ds = dremio.get_dataset(id=dataset.id)
    assert dataset.sql == ds.sql
    assert dataset.path == ds.path
    assert dataset.sql == ds.sql


def test_get_dataset_by_path(dremio: Dremio, dataset: Dataset):
    ds = dremio.get_dataset(path=dataset.path)
    assert dataset.sql == ds.sql
    assert dataset.path == ds.path
    assert dataset.sql == ds.sql


def test_get_dataset_as_catalog_object_by_id(dremio: Dremio, dataset: Dataset):
    ds = dremio.get_catalog_by_id(dataset.id)
    assert dataset.id == ds.id


def test_get_dataset_as_catalog_object_by_path(dremio: Dremio, dataset: Dataset):
    ds = dremio.get_catalog_by_path(path=dataset.path)
    assert dataset.id == ds.id


def test_run_dataset_arrow(dataset: Dataset):
    res = dataset.run("arrow")
    assert res
    assert len(res) > 0


def test_run_dataset_http(dataset: Dataset):
    res = dataset.run("http")
    assert res
    assert len(res) > 0


def test_run_dataset_to_pandas_arrow(dataset: Dataset):
    df = dataset.run("arrow").to_pandas()
    assert df is not None
    assert len(df) > 0


def test_run_dataset_to_polars_arrow(dataset: Dataset):
    df = dataset.run("arrow").to_polars()
    assert df is not None
    assert len(df) > 0


def test_run_dataset_to_pandas_http(dataset: Dataset):
    df = dataset.run("http").to_pandas()
    assert df is not None
    assert len(df) > 0


def test_run_dataset_to_polars_http(dataset: Dataset):
    df = dataset.run("http").to_polars()
    assert df is not None
    assert len(df) > 0


def test_run_dataset_to_pandas_shortcut(dataset: Dataset):
    df = dataset.run_to_pandas()
    assert df is not None
    assert len(df) > 0


def test_run_dataset_to_polars_shortcut(dataset: Dataset):
    df = dataset.run_to_polars()
    assert df is not None
    assert len(df) > 0


def test_copy_dataset(dataset: Dataset, folder: Folder):
    dataset_copy = dataset.copy(folder.path + [random_name()])
    assert dataset.id != dataset_copy.id
    assert dataset.name != dataset_copy.name
    assert dataset.sql == dataset_copy.sql
    # check if results are identical
    res1 = dataset.run()
    res2 = dataset_copy.run()
    assert res1.schema == res2.schema
    assert len(res1) == len(res2)


def test_reference_dataset(dataset: Dataset, folder: Folder):
    dataset_ref = dataset.reference(folder.path + [random_name()])
    assert dataset.id != dataset_ref.id
    assert dataset.name != dataset_ref.name
    assert type(dataset_ref.sql) == str
    assert path_to_dotted(dataset.path) in dataset_ref.sql
    # check if results are identical
    res1 = dataset.run()
    res2 = dataset_ref.run()
    assert res1.schema == res2.schema
    assert len(res1) == len(res2)


def test_wiki_get_set(dataset: Dataset):
    text = "Hallo World"
    dataset.set_wiki_text(text)
    wiki = dataset.get_wiki()
    assert wiki.text == text


def test_wiki_commit(dataset: Dataset):
    wiki = dataset.get_wiki()
    wiki.text = "Hi, I'm a Wiki"
    wiki.commit()

    wiki2 = dataset.get_wiki()
    assert wiki.text == wiki2.text
    assert wiki.version == wiki2.version


def test_refresh(dataset: Dataset):
    dataset.refresh()


def test_reflection(dataset: Dataset):
    reflections = dataset.reflections
    assert len(reflections) == 0, "There should be 0 reflections at this time"
    del reflections
    dataset.create_recommended_reflections()
    reflections = dataset.reflections
    assert (
        len(reflections) > 0
    ), "There should be some reflections after creating the recommended"
    del reflections
    dataset.delete_reflections()
    reflections = dataset.reflections
    assert (
        len(reflections) == 0
    ), "There should be 0 reflections after all should be deleted"



import re
from .converter import path_to_list, path_to_dotted


RE = r"(FROM|JOIN)\s+([\w\d\.\"]+)(?:\s*\(?\s*(\w+)\s*\)?)?"


def sql_source_parser(sql: str) -> list:
    matches = re.findall(RE, sql, re.IGNORECASE)
    return [m[1] for m in matches]


def dependencies_list_from_sql(sql: str) -> list[list[str]]:
    return [path_to_list(s) for s in sql_source_parser(sql) if s]


def dependencies_dotted_from_sql(sql: str) -> list[str]:
    return [path_to_dotted(s) for s in sql_source_parser(sql) if s]


def cut_scheme(url: str) -> str:
    """Cuts scheme from given url.

    Args:
        url (str): Like "http://localhost".

    Returns:
        str: "localhost"
    """
    if url.startswith("http"):
        url = url[url.find("//") + 2 :]
    if "/" in url:
        url = url[: url.find("/")]
    if ":" in url:
        url = url[: url.find(":")]
    return url



from dataclasses import dataclass
from requests import Response
from typing import Optional, Any


@dataclass
class DremioError(Exception):
    errorMessage: str
    moreInfo: Optional[str] = None
    status_code: Optional[int] = None
    context: Optional[Any] = None
    code: Optional[Any] = None
    details: Optional[Any] = None
    stackTrace: Optional[Any] = None

    def __post_init__(self):
        super().__init__(f"{self.status_code}: {self.errorMessage}")

    @staticmethod
    def dict_factory(x):
        exclude_fields = ("status_code",)
        return {k: v for (k, v) in x if ((v is not None) and (k not in exclude_fields))}


def exept_error_409(response: Response) -> DremioError:
    if response.status_code == 409:
        return DremioError(**response.json())
    return response.json()


@dataclass
class DremioConnectorError(Exception):
    errorMessage: str
    moreInfo: Optional[str] = None

    def __post_init__(self):
        super().__init__(f"{self.errorMessage}")


class AuthError(Exception):
    pass



import pytest
from tests.testutils.helper import random_name
from src.dremio import Dremio, DremioError, NewSpace


def test_create_get_delete_space(dremio: Dremio):
    test_space_name = random_name()
    created_space = dremio.create_catalog_item(NewSpace(test_space_name))
    space = dremio.get_catalog_by_path([test_space_name])
    assert space.id == created_space.id
    dremio.delete_catalog_item(space.id)

    with pytest.raises(DremioError):
        dremio.get_catalog_by_id(space.id)
    with pytest.raises(DremioError):
        dremio.get_catalog_by_path([test_space_name])



import re
from typing import Callable, Dict, Literal, Any, Union


class ValidationError(Exception):
    pass


class Validation:
    regex = {
        "email": r"^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$",
        "password": r"^.{8,}$",
        "username": r"^[a-zA-Z0-9_.-@]{3,}$",
        "hostname": r"^https?:\/\/(?:www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}$",
        "port": r"^[0-9]{1,5}$",
    }

    def __val(self, value, regex, error):
        # if not re.match(regex, value):
        #  raise ValidationError(error)
        return value

    def __call__(self, **kwargs: str):
        """Check if the values of the kwargs match the regex.

        Returns:
            str: if only one key is passed
            dict[str,str]: if more than one key is passed
        """
        for target, value in kwargs.items():
            self.__val(value, self.regex[target], f"{target} is invalid")
        if len(kwargs) == 1:
            return list(kwargs.values())[0]
        return kwargs

    def hostname(
        self,
        hostname: str,
        port: Union[int, None] = None,
        protocol: Union[Literal["https", "http"], None] = None,
    ) -> str:
        """Check if the hostname is valid and return it."""
        has_protocol: Callable[[str], bool] = lambda x: bool(
            re.findall(r"^(http|https):\/\/", x)
        )
        has_port: Callable[[str], bool] = lambda x: bool(re.findall(r":[0-9]{1,4}", x))

        if has_protocol(hostname):
            if protocol:
                hostname = protocol + "://" + hostname.split("://")[1]
        else:
            if protocol:
                hostname = protocol + "://" + hostname
            else:
                hostname = "https://" + hostname

        try:
            protocol = hostname.split("://")[0]  # type: ignore
        except:
            protocol = "https"

        hostname = hostname if not hostname.endswith("/") else hostname[:-1]

        if not has_port(hostname):
            if not port:
                port = 80 if protocol == "http" else 443
            if (protocol == "https" and port != 443) or (
                protocol == "http" and port != 80
            ):
                hn = hostname.split("/")[2].split(":")[0]
                hostname = hostname.replace(hn, hn + f":{port}")
        return self.__val(hostname, self.regex["hostname"], "hostname is invalid")



from typing import Optional, Any, Union
from uuid import UUID

from .accesscontrol import AccessControlList, Privilege, Role, SystemRoles, User
from ..utils import to_dict


class DremioObject:
    _dremio: Optional[Any] = None

    def to_dict(self):
        return to_dict(self)


class DremioAccessible(DremioObject):
    id: UUID

    def commit(self):
        """Commit the changes to the Dremio instance. Access role changes will be committed automatically if you don't set `lazy=True` in the methods.
        All other changes have to be committed manually with this method.
        """
        if not self._dremio:
            raise TypeError(
                "This object has no dremio instance!\nTry to commit the dataset directly in the dremio sdk"
            )
        return self._dremio.update_catalog_item(self.id, self)

    def set_access_for_role(self, role: Union[Role, SystemRoles, UUID], privileges: list[Privilege], *, lazy: bool = False):  # type: ignore
        """Add access for a role to the folder."""
        if not self._dremio:
            raise TypeError(
                "This Folder has no dremio instance!\nTry to add access for the role directly in the dremio sdk"
            )
        if self.accessControlList is None:
            self.accessControlList = AccessControlList()
        if role == "PUBLIC":
            role = self._dremio.get_role_by_name("PUBLIC")
        elif role == "ADMIN":
            role = self._dremio.get_role_by_name("ADMIN")

        role: Union[Role, UUID]
        if self.accessControlList.roles is None:
            self.accessControlList.roles = []
        role_id = role.id if isinstance(role, Role) else role
        self.accessControlList.set_access_for_role(role_id, privileges)
        if not lazy:
            self = self.commit()
        return self

    def remove_access_for_role(self, role: Union[Role, SystemRoles, UUID], lazy: bool = False):  # type: ignore
        """Remove access for a role to the folder."""
        if not self._dremio:
            raise TypeError(
                "This Folder has no dremio instance!\nTry to remove access for the role directly in the dremio sdk"
            )
        if self.accessControlList is None:
            return self
        if role == "PUBLIC":
            role = self._dremio.get_role_by_name("PUBLIC")
        elif role == "ADMIN":
            role = self._dremio.get_role_by_name("ADMIN")

        role: Union[Role, UUID]
        if self.accessControlList.roles is None:
            return self
        role_id = role.id if isinstance(role, Role) else role
        self.accessControlList.remove_access_for_role(role_id)
        if not lazy:
            self = self.commit()
        return self

    def set_access_for_user(
        self, user: Union[User, UUID], privileges: list[Privilege], lazy: bool = False
    ):
        """Add access for a user to the folder."""
        if not self._dremio:
            raise TypeError(
                "This Folder has no dremio instance!\nTry to add access for the user directly in the dremio sdk"
            )
        if self.accessControlList is None:
            self.accessControlList = AccessControlList()
        if self.accessControlList.users is None:
            self.accessControlList.users = []
        user_id = user.id if isinstance(user, User) else user
        self.accessControlList.set_access_for_user(user_id, privileges)
        if not lazy:
            self = self.commit()
        return self

    def remove_access_for_user(self, user: Union[User, UUID], lazy: bool = False):
        """Remove access for a user to the folder."""
        if not self._dremio:
            raise TypeError(
                "This Folder has no dremio instance!\nTry to remove access for the user directly in the dremio sdk"
            )
        if self.accessControlList is None:
            return self
        if self.accessControlList.users is None:
            return self
        user_id = user.id if isinstance(user, User) else user
        self.accessControlList.remove_access_for_user(user_id)
        if not lazy:
            self = self.commit()
        return self



__all__ = ["_MixinDataset"]  # this is like `export ...` in typescript

import logging
import requests

from ..utils.converter import to_dict, path_to_list, path_to_dotted
from ..models import *

from . import BaseClass
from ._catalog import _MixinCatalog
from ._sql import _MixinSQL
from ._reflection import _MixinReflection


class _MixinDataset(_MixinReflection, _MixinSQL, _MixinCatalog, BaseClass):

    def get_dataset(
        self,
        path: Union[list[str], str, None] = None,
        *,
        id: Union[UUID, str, None] = None,
    ) -> Dataset:
        """Get the dataset. [learn more](https://github.com/continental/pydremio/blob/master/docs/DATASET.md)
        \nCould raise a DremioError if the dataset does not exist or you have no permissions.

        Parameters:
          path: dataset path: "A.B.C" or ["A","B","C"].
          id: dataset uuid.

        Returns:
          Dataset: dataset object.
        """
        obj = cast(Dataset, self._get_catalog_object(id=id, path=path))
        if type(obj) != Dataset:
            raise TypeError("requested resource is not a dataset")
        obj._dremio = self  # type: ignore
        return obj

    def create_dataset(
        self,
        path: Union[list[str], str],
        sql: Union[str, SQLRequest],
        type: Literal["PHYSICAL_DATASET", "VIRTUAL_DATASET"] = "VIRTUAL_DATASET",
    ) -> Dataset:
        """Create a new dataset. [learn more](https://github.com/continental/pydremio/blob/master/docs/DATASET.md)
        \nShortcut for `Dremio.create_catalog_item(NewDataset)` with type VIRTUAL_DATASET.

        Parameters:
          path: dataset path: "A.B.C" or ["A","B","C"].
          sql: sql request as string or SQLRequest object, if sql context is needed.
          type: dataset type. Default: VIRTUAL_DATASET.

        Returns:
          Dataset: dataset object.
        """
        if isinstance(sql, str):
            sql = SQLRequest(sql)
        ds = NewDataset(path_to_list(path), type, sql.sql)
        new_ds = self.create_catalog_item(ds)  #! could raise DremioError
        return self.get_dataset(id=new_ds.id)

    def delete_dataset(self, path: Union[list[str], str]) -> bool:
        """Delete the dataset. [learn more](https://github.com/continental/pydremio/blob/master/docs/DATASET.md)
        \nCould raise a TypeError if there is no dataset under this path.
        \nCould raise a DremioError if the dataset does not exist or you have no permissions.

        Parameters:
          path: dataset path: "A.B.C" or ["A","B","C"].
        """
        if isinstance(path, str):
            path = path_to_list(path)
        dataset = self.get_dataset(path)
        return self.delete_catalog_item(dataset.id)

    def get_lineage(self, id: Union[UUID, str]) -> Lineage:
        """Get the lineage."""
        url = f"{self.hostname}/api/v3/catalog/{str(id)}/graph"
        response = requests.get(url, headers=self._headers)
        self._raise_error(response)
        return cast(Lineage, response.json())

    def get_wiki(self, id: Union[UUID, str]) -> Union[Wiki, None]:
        """Get the wiki."""
        url = f"{self.hostname}/api/v3/catalog/{str(id)}/collaboration/wiki"
        response = requests.get(url, headers=self._headers)
        if self._stand_alone_404(response):
            return
        self._raise_error(response)
        obj = cast(Wiki, response.json())
        if type(obj) != Wiki:
            raise TypeError("requested resource is not a wiki")
        obj._dataset_id = id if isinstance(id, UUID) else UUID(id)
        obj._dremio = self
        return obj

    def set_wiki(
        self, id: Union[UUID, str], wiki: Wiki, auto_versioning: bool = False
    ) -> Wiki:
        """Set the wiki."""
        url = f"{self.hostname}/api/v3/catalog/{str(id)}/collaboration/wiki"
        if auto_versioning:
            old_wiki = self.get_wiki(id)
            wiki.version = old_wiki.version if old_wiki else 0
        payload = {"text": wiki.text, "version": wiki.version}
        response = requests.post(url, json=payload, headers=self._headers)
        self._raise_error(response)
        return cast(Wiki, response.json())

    def get_tags(self, id: str) -> Union[Tags, None]:
        """Get the tags."""
        url = f"{self.hostname}/api/v3/catalog/{id}/collaboration/tag"
        response = requests.get(url, headers=self._headers)
        if self._stand_alone_404(response):
            return
        self._raise_error(response)
        return cast(Tags, response.json())

    def set_tags(self, id: str, tags: Tags) -> Tags:
        """Set the tags."""
        url = f"{self.hostname}/api/v3/catalog/{id}/collaboration/tag"
        payload = to_dict(tags)
        response = requests.post(url, json=payload, headers=self._headers)
        self._raise_error(response)
        return cast(Tags, response.json())

    def copy_dataset(
        self,
        source_path: Union[list[str], str],
        target_path: Union[list[str], str],
        *,
        assume_privileges: bool = True,
        overwrite_existing: bool = False,
    ) -> Dataset:
        """Copy the dataset by path. The copy will point on the same source like the original.
        If the target path is already exists and overwrite flag is False, it will course an DremioError.

        Parameters:
          source_path: source location path of the dataset.
          target_path: target location path of the dataset.
          assume_privileges: Assume privileges. Default: True.
          overwrite_existing: Overwrites the existing folders and datasets with the same path in the target folder. Default: False


        Returns:
          Dataset: The copied dataset.
        """
        source_ds = self.get_dataset(source_path)
        if type(source_ds) != Dataset:
            raise ValueError("requested resource is not a dataset")
        source_ds.path = path_to_list(target_path)
        try:
            self.create_catalog_item(source_ds)
        except DremioError as e:
            if e.status_code == 409:
                if not overwrite_existing:
                    raise DremioError(
                        "Conflict", f"Item at {target_path} already exists", 409
                    )
                try:
                    logging.info(
                        f"{path_to_dotted(target_path)} exists and is been overwritten"
                    )
                    self.delete_dataset(target_path)
                    self.copy_dataset(
                        source_path,
                        target_path,
                        assume_privileges=assume_privileges,
                        overwrite_existing=overwrite_existing,
                    )
                except Exception as e:
                    logging.error(
                        f"Error while overwriting {path_to_dotted(target_path)}.", e
                    )
                    raise e
            else:
                raise e
        target_ds = self.get_dataset(target_path)
        if not assume_privileges and target_ds.accessControlList:
            target_ds.accessControlList.users = []
            target_ds.accessControlList.roles = []
            target_ds.commit()
            target_ds = self.get_dataset(target_ds.path)
        return target_ds

    def reference_dataset(
        self,
        source_path: Union[list[str], str],
        target_path: Union[list[str], str],
        *,
        assume_privileges: bool = True,
        overwrite_existing: bool = False,
    ) -> Dataset:
        """Reference the dataset by path. The reference will point on the original.
        If the target path is already exists and overwrite flag is False, it will course an DremioError.

        Parameters:
          source_path: source location path of the dataset.
          target_path: target location path of the dataset.
          assume_privileges: Assume privileges. Default: True.
          overwrite_existing: Overwrites the existing folders and datasets with the same path in the target folder. Default: False

        Returns:
          Dataset: The referenced dataset.
        """
        # check if there already is a dataset on the target path
        try:
            already_existing_ds = self.get_dataset(target_path)
            if not overwrite_existing:
                raise DremioError(
                    "Conflict", f"Item at {target_path} already exists", 409
                )
            already_existing_ds.delete()
        except:
            pass
        source_ds = self.get_dataset(source_path)
        reference_statement = f"SELECT * FROM {path_to_dotted(source_path)}"
        q = SQLRequest(
            f"CREATE VDS {path_to_dotted(target_path)} AS {reference_statement}"
        )
        _, job, e = self._http_query_result(q, return_exception=True)
        if not job:
            raise DremioError("Job not started", f"Error: {e}")
        if job.jobState in ["CANCELED", "FAILED"]:
            raise DremioError(
                job.errorMessage,
                f"VDS from {source_path} at {target_path} can't be created. Message: {job.cancellationReason}",
            )
        if job.jobState != "COMPLETED":
            raise DremioError(
                "Job not completed",
                f"VDS from {source_path} at {target_path} can't be created. Message: {e}",
            )
        target_ds = self.get_dataset(target_path)
        if assume_privileges:
            target_ds.accessControlList = source_ds.accessControlList
            target_ds.commit()
            target_ds = self.get_dataset(target_ds.path)
        return target_ds

    def get_reflections_from_dataset(
        self,
        path: Union[list[str], str, None] = None,
        *,
        id: Union[UUID, str, None] = None,
    ) -> list[Reflection]:
        """Loads all reflections of the given dataset.

        Args:
            path (Union[list[str], str, None], optional): Path to dataset.
            id (Union[UUID, str, None], optional): ID of dataset. Defaults to None.

        Returns:
            list[Reflection]: List of all reflections of this dataset. Empty list if there is none.
        """
        ds = self.get_dataset(path=path, id=id)
        url = f"{self.hostname}/api/v3/dataset/{ds.id}/reflection"
        response = requests.get(url, headers=self._headers)
        self._raise_error(response)
        try:
            data = response.json()["data"]
        except KeyError:
            raise KeyError("Expected data-tag in reflections response, but not found")
        return [cast(Reflection, reflection) for reflection in data]

    def recommended_reflections(
        self, dataset_id: str | UUID, type: Literal["ALL", "RAW", "AGG"] = "ALL"
    ) -> list[NewReflection]:
        """Retrieving reflection recommendations for a Dataset.

        Args:
            dataset_id (str|UUID): Id of dataset to reflect.
            type (Literal[&quot;ALL&quot;, &quot;RAW&quot;, &quot;AGG&quot;], optional): The type of reflection recommendations you want to create and retrieve. Defaults to "ALL".

        Returns:
            Reflections (list[NewReflection]): List of recommended reflections.
        """
        url = f"{self.hostname}/api/v3/dataset/{dataset_id}/reflection/recommendation/{type}/"
        response = requests.post(url, headers=self._headers)
        self._raise_error(response)
        try:
            data = response.json()["data"]
        except KeyError:
            raise KeyError("Expected data-tag in reflections response, but not found")
        return [cast(NewReflection, reflection) for reflection in data]

    # This is here bc it needs `get_dataset`
    def create_recommended_reflections(
        self, dataset_id: str | UUID, type: Literal["ALL", "RAW", "AGG"] = "ALL"
    ) -> list[Reflection]:
        """Creating and Retrieving Reflection Recommendations for a Dataset.

        Args:
            dataset_id (str|UUID): Id of dataset to reflect.
            type (Literal[&quot;ALL&quot;, &quot;RAW&quot;, &quot;AGG&quot;], optional): The type of reflection recommendations you want to create and retrieve. Defaults to "ALL".

        Returns:
            Reflections (list[Reflection]): List of all reflections of the dataset.
        """
        ds = self.get_dataset(id=dataset_id)
        recommanded_reflections = self.recommended_reflections(
            dataset_id=dataset_id, type=type
        )
        refs: list[Reflection] = []
        for ref in recommanded_reflections:
            name = f"{ds.path[-1]}_{ref.type}_{str(ds.id)[:4]}"
            r = self.create_reflection(dataset_id=dataset_id, name=name, reflection=ref)
            refs.append(r)
        return refs

    def refresh_dataset(self, path: Union[list[str], str]) -> None:
        """Refresh dataset reflection"""
        ds = self.get_dataset(path)
        self._refresh_catalog(ds.id)

    def refresh_dataset_metadata(self, path: Union[list[str], str]):
        """Refresh metadata of physical datasets.

        Args:
            path (Union[list[str], str]): Path to physical dataset.
        """
        sql = f"ALTER TABLE {path_to_dotted(path)} REFRESH METADATA;"
        self._http_query_result(sql)
